import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.DialogBoxClass;
import win32frame.PushButton;
import win32frame.TextBox;
import util.BitMap;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import Spin;
import MyButton ;
import ColorSelectDialog;

import HairDB;

COLORREF RGB(uint r, uint g, uint b) {
  return cast(COLORREF)(r | (g << 8) | (b << 16));
}
COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

enum CMD{
  DENSITY = 0x100,
  BZSTEP,
  FILTER,
  DIFFUSE,
  SPECULAR,
  PHONG,
  PHONGSIZE,
  RADIUS_S,
  RADIUS_E,
  SELCOL,
  OK,
  CANCEL
} ;

struct SpinList{
  Spin   wobj ;
  string label ;
  HMENU  cmd ;
  string fmt ;
  double cur ;
  double min ;
  double max ;
  double step ;

  void set(string label, HMENU cmd, string fmt, double cur, double min, double max, double step){
    this.label=label ;
    this.cmd  =cmd ;
    this.fmt  =fmt ;
    this.cur  =cur ;
    this.min  =min ;
    this.max  =max ;
    this.step =step ;
  }
} ;

class ExportPOVDialog : DialogBoxClass {
  SpinList[]  spinlist ;
  MyButton    btncol ;

  uint        color ;
  int         img_w=80 ;
  int         img_h=40 ;
  
  PushButton  buttonOK  ;
  PushButton  buttonCAN ;
  HFONT       hfont  ;

  POVParam    param ;
  int         dlgstate  ;

  const int   defaultWidth =200 ;
  const int   defaultHeight=300 ;
  const DWORD defaultStyle = WS_VISIBLE | WS_SYSMENU ;

  this(string name, string ClassName) {
    UnregisterClass( toUTFz!(wchar*)(ClassName), GetModuleHandle(null) ) ;
    winc.cbSize         = WNDCLASSEX.sizeof;
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = null ;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
  }

  int Create( int x, int y, HWND phwnd, HMENU hmenu){
    return( create(defaultStyle,x,y,defaultWidth,defaultHeight,phwnd,hmenu) ) ;
  }
  
  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont);
      break ;
    case WM_CREATE:
      hfont = CreateFont(
			 12 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
			 DEFAULT_CHARSET, OUT_DEFAULT_PRECIS ,
			 CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY , 
			 FIXED_PITCH, null
			 );

      int spinnum=9 ;
      spinlist = new SpinList[spinnum] ;

      spinlist[0].set("Density",     cast(HMENU)(CMD.DENSITY),  "%.0f", param.density,     0,   100000,     10     ) ;
      spinlist[1].set("BezierStep",  cast(HMENU)(CMD.BZSTEP),   "%.0f", param.bezier_step, 8,       32,      1     ) ;
      spinlist[2].set("Filter",      cast(HMENU)(CMD.FILTER),   "%.3f", param.f,           0,        1.0,    0.01  ) ;
      spinlist[3].set("Diffuse",     cast(HMENU)(CMD.DIFFUSE),  "%.3f", param.diffuse,     0,        1.0,    0.01  ) ;
      spinlist[4].set("Specular",    cast(HMENU)(CMD.SPECULAR), "%.3f", param.specular,    0,        1.0,    0.01  ) ;
      spinlist[5].set("Phong",       cast(HMENU)(CMD.PHONG),    "%.3f", param.phong,       0,        1.0,    0.01  ) ;
      spinlist[6].set("PhongSize",   cast(HMENU)(CMD.PHONGSIZE),"%.2f", param.phong_size,  1,      100.0,    1.00  ) ;
      spinlist[7].set("StartRadius", cast(HMENU)(CMD.RADIUS_S), "%.4f", param.radius_s,    0.0000,   0.1000, 0.0001) ;
      spinlist[8].set("EndRadius",   cast(HMENU)(CMD.RADIUS_E), "%.4f", param.radius_e,    0.0000,   0.1000, 0.0001) ;
      
      foreach( i, ref spin ; spinlist ){
	spin.wobj = new Spin("",spin.label) ;
	spin.wobj.setStyle( WS_CHILD|WS_CLIPSIBLINGS|WS_CLIPCHILDREN|WS_VISIBLE|WS_TABSTOP ) ;
	spin.wobj.setPosition(100, 8+cast(int)(i)*20,80,20) ;
	spin.wobj.setTextFormat(spin.fmt) ;
	spin.wobj.setInfo( spin.cur, spin.min, spin.max, spin.step ) ;
	addObject(spin.wobj, spin.cmd, false, false) ;
	spin.wobj.setTextFont(hfont) ;
      }

      color = RGBA(cast(uint)(round(param.r*255)),cast(uint)(round(param.g*255)),cast(uint)(round(param.b*255)),cast(uint)(round((1.0-param.t)*255))) ;
      auto colimg = new BitMap() ;
      colimg.setImage(makeColorButtonFace(color,img_w,img_h),img_w,img_h,false,false) ;
      
      btncol= new MyButton("ColorSelect") ;
      btncol.setBitMap(colimg) ;
      btncol.addStyle(WS_TABSTOP) ;
      btncol.setPosition(100,190,img_w,img_h) ;
      addObject( btncol, cast(HMENU)(CMD.SELCOL) ) ;
      btncol.setFont(hfont) ;

      buttonOK= new PushButton("OK") ;
      buttonOK.addStyle(WS_TABSTOP) ;
      buttonOK.setPosition(winW-184,winH-32,80,20) ;
      addObject( buttonOK, cast(HMENU)(CMD.OK) ) ;
      buttonOK.setFont(hfont) ;

      buttonCAN= new PushButton("Cancel") ;
      buttonCAN.addStyle(WS_TABSTOP) ;
      buttonCAN.setPosition(winW-92,winH-32,80,20) ;
      addObject( buttonCAN, cast(HMENU)(CMD.CANCEL) ) ;
      buttonCAN.setFont(hfont) ;

      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC hdc = BeginPaint(hwnd, &ps);
      SetBkMode(hdc , TRANSPARENT);
      foreach( i, spin ; spinlist ){
	SIZE labelsize ;
	wchar[] label=(toUTF16(spin.label)).dup ;

	SelectObject(hdc,hfont) ;
	GetTextExtentPoint32( hdc, cast(wchar*)(label), cast(int)(label.length), &labelsize);
	TextOut(hdc,
		16,8+4+cast(int)(i)*20,
		cast(wchar*)(label), cast(int)(label.length)) ;
      }
      {
	SIZE labelsize ;
	wchar[] label=(toUTF16("Color")).dup ;
	SelectObject(hdc,hfont) ;
	GetTextExtentPoint32( hdc, cast(wchar*)(label), cast(int)(label.length), &labelsize);
	TextOut(hdc,
		16,2+190,
		cast(wchar*)(label), cast(int)(label.length)) ;
      }
	
      EndPaint(hwnd, &ps);
      break ;
      
    case WM_SIZE:
      break ;
      
    case WM_CLOSE:
      setResult(0) ;
      DestroyWindow(hwnd);
      break ;
      
    case WM_COMMAND:
      dlgstate=LOWORD(wp) ;
      switch( dlgstate ){
      case CMD.SELCOL:
	RECT wrect ;
        GetWindowRect(this.getMainWinHWND(),&wrect) ;
        auto colsel = new ColorSelectDialog("","colsel") ;
	colsel.setColor(color) ;
	colsel.enableAlpha(true) ;
	colsel.Create( wrect.left+120,wrect.top+120,hwnd,null ) ;
	if( colsel.getDlogResult()==1 ){
          color = colsel.getColor() ;
	  auto colimg = new BitMap() ;
	  colimg.setImage(makeColorButtonFace(color,img_w,img_h),img_w,img_h,false,false) ;
	  btncol.setBitMap(colimg) ;
	  RedrawWindow( btncol.getHWND(), null, null, RDW_INVALIDATE ) ;
        }
        break ;

      case CMD.CANCEL:
        setResult(0) ;
        DestroyWindow(hwnd);
        break ;
      case CMD.OK:
        setResult(1) ;
        DestroyWindow(hwnd);
        break ;
	
      default :
	break ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int setDefaultPOVparam(POVParam param){
    this.param=param ;
    return(0) ;
  }
  
  POVParam getPOVparam(){
    param.r=cast(double)((color    )&0x000000ff)/255.0 ;
    param.g=cast(double)((color>>8 )&0x000000ff)/255.0 ;
    param.b=cast(double)((color>>16)&0x000000ff)/255.0 ;
    param.t=1.0 - cast(double)((color>>24)&0x000000ff)/255.0 ;
    param.density    = spinlist[0].wobj.getValue() ;
    param.bezier_step= spinlist[1].wobj.getValue() ;
    param.f          = spinlist[2].wobj.getValue() ;
    param.diffuse    = spinlist[3].wobj.getValue() ;
    param.specular   = spinlist[4].wobj.getValue() ;
    param.phong      = spinlist[5].wobj.getValue() ;
    param.phong_size = spinlist[6].wobj.getValue() ;
    param.radius_s   = spinlist[7].wobj.getValue() ;
    param.radius_e   = spinlist[8].wobj.getValue() ;
    return(param) ;
  }

  uint[] makeColorButtonFace(uint col, int w, int h){
    uint[] img = new uint[w*h] ;
    for( int y=0 ; y<h ; y++ ){
      for( int x=0 ; x<w ; x++ ){
	int bgv = ((x/8+y/8)%2)==0 ? 128 : 70 ;
	int a = (col>>24) & 0x000000ff ;
	int r = (col    ) & 0x000000ff ;
	int g = (col>>8 ) & 0x000000ff ;
	int b = (col>>16) & 0x000000ff ;
	r = (r*a + bgv*(255-a)) /255 ;
	g = (g*a + bgv*(255-a)) /255 ;
	b = (b*a + bgv*(255-a)) /255 ;
        img[y*w+x]=RGBA(b,g,r,255) ;
      }
    }
    return(img) ;
  }
    
}
  
