import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.utf;
import std.format;

import Vec3D ;
import HairDB ;

class FileHair{
  HairRootDB  hrDB ;
  HairGuideDB hgDB ;

  this(){
  }

  HairRootDB getHairRootDB() {
    return(hrDB) ;
  }
  
  HairGuideDB getHairGuideDB() {
    return(hgDB) ;
  }
    
  int write(string fname, HairRootDB  hrDB, HairGuideDB hgDB){
    BufferedFile outfile = new BufferedFile(fname, FileMode.OutNew) ;
    vec3D[]  HRvtx = hrDB.getHRvtx() ;
    face3D[] HRface= hrDB.getHRface() ;
    outfile.writeString( format("PN %d\n",HRvtx.length)  ) ;
    outfile.writeString( format("FN %d\n",HRface.length) ) ;
    for( int i=0 ; i<HRvtx.length ; i++ ){
      outfile.writeString( format("P %d %f %f %f\n",i,HRvtx[i].x,HRvtx[i].y,HRvtx[i].z) ) ;
    }
    for( int i=0 ; i<HRface.length ; i++ ){
      outfile.writeString( format("F %d %d %d %d\n",i,HRface[i].p0,HRface[i].p1,HRface[i].p2) ) ;
    }

    HairGuide[] hglist = hgDB.getHGlist() ;
    for( int i=0 ; i<hglist.length ; i++ ){
      outfile.writeString( format("HG %d %d %d %d",hglist[i].type, hglist[i].gid, hglist[i].HRid, hglist[i].plist.length) ) ;
      for( int j=0 ; j<hglist[i].plist.length ; j++ ){
	outfile.writeString( format(" %f %f %f", hglist[i].plist[j].x, hglist[i].plist[j].y, hglist[i].plist[j].z) ) ;
      }
      outfile.writeString("\n") ;
    }
    outfile.close() ;

    return(0) ;
  }

  int read(string fname, int maxgid){
    hrDB = new HairRootDB() ;
    hgDB = new HairGuideDB(maxgid) ;
    vec3D[]  HRvtx  ;
    vec3D[]  HRnml  ;
    face3D[] HRface ;
    int vcount=0 ;
    int fcount=0 ;
    
    try{
      BufferedFile infile = new BufferedFile(fname) ;
      while(-1){
	string line=cast(string)infile.readLine() ;
	string[] token ;
	if( line is null ) line="" ;
	if( line.length==0 ){
	  if( infile.eof()==true ) break ;
	  continue ;
	}
	token = split(line) ;
	switch( token[0] ){
	case "PN" :
	  if( token.length==2 ){
	    int vnum=to!(int)(token[1]) ;
	    if( vnum>0 ) HRvtx.length=vnum ;
	  }
	  break ;
	case "FN" :
	  if( token.length==2 ){
	    int fnum=to!(int)(token[1]) ;
	    if( fnum>0 ) HRface.length=fnum ;
	  }
	  break ;
	case "P" :
	  if( token.length==5 ){
	    int vno=to!(int)(token[1]) ;
	    if( vno!=vcount ) throw new Exception("vertex number is not match") ;
	    if( vno<HRvtx.length && vno>=0 ){
	      HRvtx[vcount].x=to!(double)(token[2]) ;
	      HRvtx[vcount].y=to!(double)(token[3]) ;
	      HRvtx[vcount].z=to!(double)(token[4]) ;
	      HRvtx[vcount].id=vno ;
	      vcount++ ;
	    }
	  }
	  break ;
	case "F" :
	  if( token.length==5 ){
	    int fno=to!(int)(token[1]) ;
	    if( fno!=fcount ) throw new Exception("face number is not match") ;
	    if( fno<HRface.length && fno>=0 ){
	      HRface[fcount].p0=to!(int)(token[2]) ;
	      HRface[fcount].p1=to!(int)(token[3]) ;
	      HRface[fcount].p2=to!(int)(token[4]) ;
	      HRface[fcount].id=fno ;
	      fcount++ ;
	    }
	  }
	  break ;
	case "HG" :
	  if( token.length==26 ){
	    int  gid=to!(int)(token[2]) ;
	    int hrid=to!(int)(token[3]) ;
	    if( hrid>=vcount ) throw new Exception("Illegal HairRootID") ;
	    if( to!(int)(token[4])==7 ){
	      vec3D[]  plist  ; plist.length=7 ;
	      plist[0]=HRvtx[hrid] ;
	      for( int i=1 ; i<7 ; i++ ){
		plist[i].x=to!(double)(token[5+i*3+0]) ;
		plist[i].y=to!(double)(token[5+i*3+1]) ;
		plist[i].z=to!(double)(token[5+i*3+2]) ;
	      }
	      hgDB.add( hrid, gid, false, plist) ;
	    }
	  }
	  break ;
	default :
	  break ;
	}
      }
      infile.close() ;
    }catch(Exception e){
      throw e ;
    }
    
    hrDB.set(HRvtx, HRface) ;
    hgDB.setHRnmlList(hrDB.getHRnml()) ;
    
    return(0) ;
  }
}
