import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.regex;
import std.file;

import Vec3D;
import HairDB ;

class FileOBJ
{
  Regex!(char) getfstwd ;
  Regex!(char) obj_reg_tbl[string] ;
  Regex!(char) mtl_reg_tbl[string] ;
  int delegate(string, string) obj_act_tbl[string] ;
  int delegate(string, string) mtl_act_tbl[string] ;

  HairRootDB  hrDB ;
  HairGuideDB hgDB ;

  vec3D[]  HRvtx  ;
  vec3D[]  HRnml  ;
  face3D[] HRface ;
  int      vcount ;
  int      fcount ;

  double   scale  ;
  
  this(){
    string nume = "[-]*[0-9]+\\.[0-9]+[eE]{1}[-+]?[0-9]+" ;
    string numd = "[-]*[0-9]+\\.[0-9]+" ;
    string numi = "[-]*[0-9]+" ;
    string num  ;
    
    num = "(" ~ nume ~ "|" ~ numd ~ "|" ~ numi ~ ")" ;

    obj_reg_tbl["mtllib"] = regex( "^mtllib\\s+(\\S+)"                            ,"") ;
    obj_reg_tbl["o"     ] = regex( "^o\\s+(\\S+)"                                 ,"") ;
    obj_reg_tbl["s"     ] = regex( "^s\\s+(\\S+)"                                 ,"") ;
    obj_reg_tbl["v"     ] = regex( "^v\\s+"  ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    obj_reg_tbl["vn"    ] = regex( "^vn\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    obj_reg_tbl["vt"    ] = regex( "^vt\\s+" ~ num ~ "\\s+" ~ num                 ,"") ;
    obj_reg_tbl["g"     ] = regex( "^g\\s+(\\S+)"                                 ,"") ;
    obj_reg_tbl["usemtl"] = regex( "^usemtl\\s+(\\S+)"                            ,"") ;
    obj_reg_tbl["f"     ] = regex( "^f\\s+([0-9]*)/([0-9]*)/([0-9]*)\\s+([0-9]*)/([0-9]*)/([0-9]*)\\s+([0-9]*)/([0-9]*)/([0-9]*)","") ;
    obj_reg_tbl["#"     ] = regex( "^#"                                           ,"") ;

    obj_act_tbl["mtllib"] = &nop ;
    obj_act_tbl["o"     ] = &nop ;
    obj_act_tbl["v"     ] = &add_vertex ;
    obj_act_tbl["vn"    ] = &nop ;
    obj_act_tbl["vt"    ] = &nop ;
    obj_act_tbl["usemtl"] = &nop ;
    obj_act_tbl["f"     ] = &add_face   ;

    mtl_reg_tbl["newmtl"] = regex( "^newmtl\\s+(\\S+)"                            ,"") ;
    mtl_reg_tbl["Ns"    ] = regex( "^Ns\\s+" ~ num                                ,"") ;
    mtl_reg_tbl["d"     ] = regex( "^d\\s+" ~ num                                 ,"") ;
    mtl_reg_tbl["illum" ] = regex( "^illum\\s+" ~ num                             ,"") ;
    mtl_reg_tbl["Kd"    ] = regex( "^Kd\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ka"    ] = regex( "^Ka\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ks"    ] = regex( "^Ks\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ke"    ] = regex( "^Ke\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["map_Kd"] = regex( "^map_Kd\\s+(\\S+)"                            ,"") ;
    
    mtl_act_tbl["newmtl"] = &nop ;
    mtl_act_tbl["Ns"    ] = &nop ;
    mtl_act_tbl["Kd"    ] = &nop ;
    mtl_act_tbl["Ka"    ] = &nop ;
    mtl_act_tbl["Ks"    ] = &nop ;
    mtl_act_tbl["map_Kd"] = &nop ;

    getfstwd = regex("^(\\S+)","") ;

  }

  HairRootDB getHairRootDB() {
    return(hrDB) ;
  }
  
  HairGuideDB getHairGuideDB() {
    return(hgDB) ;
  }
    
  int read(string fname, int maxgid, double scale=1.0){
    HRvtx.length =200 ;
    HRface.length=200 ;
    vcount = 0 ;
    fcount = 0 ;

    this.scale=scale ;

    BufferedFile infile = new BufferedFile(fname) ;
    
    while(-1){
      string line=cast(string)infile.readLine() ;
      if( line is null ) line="" ;
      if( line.length==0 ){
	if( infile.eof()==true ) break ;
	continue ;
      }
      
      auto token = match(line, getfstwd) ;
      if( token.empty==false ){
	auto capt = token.captures ;
	string fstword = capt[1] ;
	if( (match(fstword, obj_reg_tbl["#"])).empty==false ) continue ;

	if( (fstword in obj_reg_tbl) == null ){
	  //writef("keyword %s not found\n", fstword) ;
	  continue ;
	}
	if( (fstword in obj_act_tbl) != null ){
	  obj_act_tbl[fstword](fstword,line) ;
	}else{
	  continue ;
	}
      }
    }

    HRvtx.length =vcount ;
    HRface.length=fcount ;
    
    hrDB = new HairRootDB() ;
    hgDB = new HairGuideDB(maxgid) ;
    
    hrDB.set(HRvtx, HRface) ;
    hgDB.setHRnmlList(hrDB.getHRnml()) ;

    infile.close() ;

    return(0) ;
  }

  private int add_vertex(string act, string line){
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      if( vcount>=HRvtx.length ){
	HRvtx.length = HRvtx.length*2 ;
      }
      auto capt = token.captures ;
      HRvtx[vcount].set( std.conv.to!(double)(capt[1]) * scale,
			 std.conv.to!(double)(capt[2]) * scale,
			 std.conv.to!(double)(capt[3]) * scale ) ;
      vcount++ ;
    }
    return(0) ;
  }

  private int add_face(string act, string line){
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      if( fcount>=HRface.length ){
	HRface.length = HRface.length*2 ;
      }
      auto capt = token.captures ;
      if(  capt[1].length!=0 &&  capt[4].length!=0 &&  capt[7].length!=0 ){
	HRface[fcount].set(std.conv.to!(int)(capt[7]) -1,
			   std.conv.to!(int)(capt[4]) -1,
			   std.conv.to!(int)(capt[1]) -1 ) ;
	fcount++ ;
      }
    }
    return(0) ;
  }

  private int nop(string act, string line){
    return(0) ;
  }
  
}

