import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.utf;
import std.format;
import mt19937ar;

import Vec3D;
import HairDB ;

struct POVParam{
  double bezier_step =     16 ;
  double density     =  25000 ;
  double r           =0.50000 ;
  double g           =0.20000 ;
  double b           =0.00000 ;
  double f           =0.00000 ;
  double t           =0.00000 ;
  double diffuse     =0.60000 ;
  double specular    =0.00000 ;
  double phong       =0.80000 ;
  double phong_size  =7.0     ;
  double radius_s    =0.0009  ;
  double radius_e    =0.0007  ;
} ;
  
class FilePOV
{
  
  this(){
  }

  int write(string fname, HairRootDB  hrDB, HairGuideDB hgDB, POVParam param){
    BufferedFile outfile = new BufferedFile(fname, FileMode.OutNew) ;
    vec3D[]  HRvtx = hrDB.getHRvtx() ;
    face3D[] HRface= hrDB.getHRface() ;
    vec3D[]  plist = new vec3D[POINT_NUM] ;

    write_hair_param(outfile,param) ;
    write_macro(outfile, param) ;
    write_obj_header(outfile, param) ;
    
    HairGuide[] hglist = hgDB.getHGlist() ;
    GIDSTATE[] GIDstatlist = hgDB.getGIDstatlist() ;

    foreach( gid, gidstat ; GIDstatlist ){
      if( gidstat.hide==true ) continue ;
      foreach( hg ; hglist ){
	if( hg.gid==gid ){
	  write_object(outfile, hg.plist, param) ;
	}
      }
      foreach( f ; HRface ){
	int fp0_hgid=-1 ;
	int fp1_hgid=-1 ;
	int fp2_hgid=-1 ;
	foreach( hgid, hg ; hglist ){
	  if( hg.gid!=gid ) continue ;
	  if( fp0_hgid==-1 && HRvtx[f.p0].id==hg.HRid ){ fp0_hgid=cast(int)(hgid) ; }
	  if( fp1_hgid==-1 && HRvtx[f.p1].id==hg.HRid ){ fp1_hgid=cast(int)(hgid) ; }
	  if( fp2_hgid==-1 && HRvtx[f.p2].id==hg.HRid ){ fp2_hgid=cast(int)(hgid) ; }
	}
	if( fp0_hgid>=0 && fp1_hgid>=0 && fp2_hgid>=0 ){ //三点にヘアガイドが存在する
	  vec3D  o0 ; o0.set(0.0,0.0,0.0) ;
	  vec3D  v01=HRvtx[f.p1] - HRvtx[f.p0] ;
	  vec3D  v02=HRvtx[f.p2] - HRvtx[f.p0] ;
	  double sa=(v01.cross(v02)).length()/2.0 ;
	  int gencount = cast(int)(param.density*sa) ;
	  for( int i=0 ; i<gencount ; i++ ){
	    vec3D px = (v01 * genrand_real1()) + (v02 * genrand_real1()) ;
	    if( px.chkinside(o0,v01,v02)==false ){
	      // 乱数に偏りがあると無限ループする可能性がゼロではない
	      i-- ;
	      continue ;
	    }
	    double s1=((-px).cross(v02-px)).length()/2.0 ;
	    double s2=((-px).cross(v01-px)).length()/2.0 ;
	    for( int j=0 ; j<hglist[fp0_hgid].plist.length ; j++ ){
	      vec3D p0,pa,pb ;
	      p0=hglist[fp0_hgid].plist[j] * ((sa-s1-s2)/sa) ;
	      pa=hglist[fp1_hgid].plist[j] * (s1/sa) ;
	      pb=hglist[fp2_hgid].plist[j] * (s2/sa) ;
	      plist[j]=p0+pa+pb ;
	    }
	    write_object(outfile, plist, param) ;
	  }
	}
      }
    }

    write_obj_footer(outfile,param) ;
    outfile.close() ;

    return(0) ;
  }

  private int write_object(BufferedFile outfile, vec3D[] plist, POVParam param){
    bool same1=(plist[0]==plist[1] && plist[2]==plist[3] && (plist[0]-plist[3]).length()==0.0) ;
    bool same2=(plist[3]==plist[4] && plist[5]==plist[6] && (plist[3]-plist[6]).length()==0.0) ;
    if( same1==true || same2==true ) return(-1) ; //長さが0のシリンダを生成するケースを除外
    
    if( plist.length==1 ){
      outfile.writeString(format("      BezierCurve1( ")) ;
    }else{
      outfile.writeString(format("      BezierCurve2( ")) ;
    }
    for( int i=0 ; i<plist.length ; i++ ){
      outfile.writeString(format("<%f,%f,%f>,",-plist[i].x,plist[i].y,plist[i].z)) ; //??何故かX軸極性が逆に？
    }
    outfile.writeString(format("HM_RS,HM_RE")) ;
    outfile.writeString(format(")\n")) ;
    return(0) ;
  }

  private int write_hair_param(BufferedFile outfile, POVParam param){
    outfile.writeString(format("#declare HM_RS=%f;\n",param.radius_s)) ;
    outfile.writeString(format("#declare HM_RE=%f;\n",param.radius_e)) ;
    outfile.writeString(format("#declare HM_color =\n")) ;
    outfile.writeString(format("texture{\n")) ;
    outfile.writeString(format("  pigment{ rgbft <%f, %f, %f, %f, %f>}\n",
			       param.r, param.g, param.b, param.f, param.t)) ;
    outfile.writeString(format("  finish{\n")) ;
    outfile.writeString(format("    //ambient rgb <1.00000, 1.00000, 1.00000>\n")) ;
    outfile.writeString(format("    diffuse %f\n",param.diffuse)) ;
    outfile.writeString(format("    brilliance 1.00000\n")) ;
    outfile.writeString(format("    specular %f roughness 0.0040000\n",param.specular)) ;
    outfile.writeString(format("    phong %f phong_size %f\n",param.phong, param.phong_size)) ;
    outfile.writeString("  }\n"
			"}\n"
			"\n") ;
    return(0) ;
  }
    
  private int write_macro(BufferedFile outfile, POVParam param){
    outfile.writeString("#macro BezierCurve1(Origin, Ctrl1, Ctrl2, End, RadiusS, RadiusE)\n") ;
    outfile.writeString(format("  #local Step = 1/%f;\n",param.bezier_step)) ;
    outfile.writeString("  #local Rdiff= RadiusS - RadiusE;\n"
			"  union{\n"
			"    #local Pp = Origin;\n"
			"    #local T=Step;\n"
			"    #local TotalLength=0;\n"
			"    #while (T <= 1+Step/2)\n"
			"      #local U = 1-T;\n"
			"      #local P = U*U*U*Origin + 3*U*U*T*Ctrl1 + 3*U*T*T*Ctrl2 + T*T*T*End;\n"
			"      #local PpP= P-Pp;\n"
			"      #local TotalLength=TotalLength+vlength(PpP);\n"
			"      #local Pp = P;\n"
			"      #local T = T+Step;\n"
			"    #end\n"
			"\n"
			"    #local Pp = Origin;\n"
			"    #local T=Step;\n"
			"    #local Lp=0;\n"
			"    #while (T <= 1+Step/2)\n"
			"      #local U = 1-T;\n"
			"      #local P = U*U*U*Origin + 3*U*U*T*Ctrl1 + 3*U*T*T*Ctrl2 + T*T*T*End;\n"
			"      #local Ro= RadiusS - (Rdiff/TotalLength*Lp);\n"
			"      #local PpP= P-Pp;\n"
			"      #local Lp= Lp + vlength(PpP);\n"
			"      #local Re= RadiusS - (Rdiff/TotalLength*Lp);\n"
			"      cone{Pp, Ro, P, Re}\n"
			"      #local Pp = P;\n"
			"      #local T = T+Step;\n"
			"    #end\n"
			"  }\n"
			"#end\n"
			"\n") ;
    outfile.writeString("#macro BezierCurve2(Origin, Ctrl1, Ctrl2, Center, Ctrl3, Ctrl4, End, RadiusS, RadiusE)\n") ;
    outfile.writeString(format("  #local Step = 1/%f;\n",param.bezier_step)) ;
    outfile.writeString("  #local Rdiff= RadiusS - RadiusE;\n"
			"  union{\n"
			"    #local Pp = Origin;\n"
			"    #local T=Step;\n"
			"    #local TotalLength=0;\n"
			"    #while (T <= 1+Step/2)\n"
			"      #local U = 1-T;\n"
			"      #local P = U*U*U*Origin + 3*U*U*T*Ctrl1 + 3*U*T*T*Ctrl2 + T*T*T*Center;\n"
			"      #local PpP= P-Pp;\n"
			"      #local TotalLength=TotalLength+vlength(PpP);\n"
			"      #local Pp = P;\n"
			"      #local T = T+Step;\n"
			"    #end\n"
			"    #local Pp = Center;\n"
			"    #local T=Step;\n"
			"    #while (T <= 1+Step/2)\n"
			"      #local U = 1-T;\n"
			"      #local P = U*U*U*Center + 3*U*U*T*Ctrl3 + 3*U*T*T*Ctrl4 + T*T*T*End;\n"
			"      #local PpP= P-Pp;\n"
			"      #local TotalLength=TotalLength+vlength(PpP);\n"
			"      #local Pp = P;\n"
			"      #local T = T+Step;\n"
			"    #end\n"
			"\n"
			"    #local Pp = Origin;\n"
			"    #local T=Step;\n"
			"    #local Lp=0;\n"
			"    #while (T <= 1+Step/2)\n"
			"      #local U = 1-T;\n"
			"      #local P = U*U*U*Origin + 3*U*U*T*Ctrl1 + 3*U*T*T*Ctrl2 + T*T*T*Center;\n"
			"      #local Ro= RadiusS - (Rdiff/TotalLength*Lp);\n"
			"      #local PpP= P-Pp;\n"
			"      #local Lp= Lp + vlength(PpP);\n"
			"      #local Re= RadiusS - (Rdiff/TotalLength*Lp);\n"
			"      cone{Pp, Ro, P, Re}\n"
			"      #local Pp = P;\n"
			"      #local T = T+Step;\n"
			"    #end\n"
			"    #local Pp = Center;\n"
			"    #local T=Step;\n"
			"    #while (T <= 1+Step/2)\n"
			"      #local U = 1-T;\n"
			"      #local P = U*U*U*Center + 3*U*U*T*Ctrl3 + 3*U*T*T*Ctrl4 + T*T*T*End;\n"
			"      #local Ro= RadiusS - (Rdiff/TotalLength*Lp);\n"
			"      #local PpP= P-Pp;\n"
			"      #local Lp= Lp + vlength(PpP);\n"
			"      #local Re= RadiusS - (Rdiff/TotalLength*Lp);\n"
			"      cone{Pp, Ro, P, Re}\n"
			"      #local Pp = P;\n"
			"      #local T = T+Step;\n"
			"    #end\n"
			"  }\n"
			"#end\n"
			"\n") ;
    return(0) ;
  }

  private int write_obj_header(BufferedFile outfile, POVParam param){
    outfile.writeString("object{\n"
			"   union{\n") ;
    return(0) ;
  }

  private int write_obj_footer(BufferedFile outfile, POVParam param){
    outfile.writeString("   }\n"
			"   texture{\n"
			"     HM_color\n"
			"   }\n"
			"}\n") ;
    return(0) ;
  }

}

