import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

import GidPanel ;

import HairDB ;
import HairGuideDB ;
import OpManager ;

COLORREF RGB(uint r, uint g, uint b) {
  return cast(COLORREF)(r | (g << 8) | (b << 16));
}

const int GIDPANEL_HEIGHT = 32 ;

class GidListPanel : WindowObject {
  HairDB       hairdb ;
  OpManager    opman ;

  GidPanel[]   gplist   ;
  int          curpanel=-1  ;
  
  SCROLLINFO   rollinfo_V ;
  

  this(string name) {
    winc.style          = 0;
    this.setStyle(WS_CHILD | WS_TABSTOP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)("GidListPanel") ;
    super(name) ;
    this.registwinc() ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      break ;
      
    case WM_CREATE :
      rollinfo_V.cbSize = SCROLLINFO.sizeof ;
      rollinfo_V.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE | SIF_DISABLENOSCROLL ;
      rollinfo_V.nPos   = 0          ;
      rollinfo_V.nPage  = winH-1     ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = winH-1     ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      break ;
      
    case WM_PAINT:
      HDC hdc;
      PAINTSTRUCT ps ;
      HBRUSH BCOLBG ;
      
      BCOLBG = CreateSolidBrush(RGB(64,64,64)) ;
      
      hdc = BeginPaint(hwnd, &ps);
      SelectObject(hdc, BCOLBG) ;
      Rectangle(hdc, 0,0,winW,winH) ;
      EndPaint(hwnd, &ps);

      SCROLLINFO  slinfo_V ;
      slinfo_V.cbSize = SCROLLINFO.sizeof ; slinfo_V.fMask = SIF_POS ;
      GetScrollInfo( hwnd, SB_VERT, &slinfo_V) ;
      
      for( int i=0 ; i<gplist.length ; i++ ){
	int  ypos   = (cast(int)(gplist.length)-1-i)*(GIDPANEL_HEIGHT+2)+2-slinfo_V.nPos ;
	bool redraw = (ypos>=0 || (ypos+GIDPANEL_HEIGHT)<=winH) ? true : false ;
	gplist[i].movePanel(0,(cast(int)(gplist.length)-1-i)*(GIDPANEL_HEIGHT+2)+2-slinfo_V.nPos,
			    winW,GIDPANEL_HEIGHT,redraw) ;
	if( i==curpanel ){
	  gplist[i].setSelected(true) ;
	}else{
	  gplist[i].setSelected(false) ;
	}
      }
      
      DeleteObject(BCOLBG ) ;
      break ;

    case WM_SIZE:
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL ;
      rollinfo_V.nPage  = winH-1     ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = cast(int)(gplist.length)*(GIDPANEL_HEIGHT+2) ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;

      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_WINDOWPOSCHANGING :
      //Zオーダーの変更があった場合にスクロールバーを再描画しないとダメ
      rollinfo_V.fMask  = SIF_POS | SIF_DISABLENOSCROLL ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      break ;
      
    case WM_VSCROLL:
      bool changed=true ;
      int  rollstep=20 ;
      
      switch( LOWORD(wp) ){
      case SB_LINEUP :
	rollinfo_V.nPos -=rollstep ;
	if( rollinfo_V.nPos < 0 ){
	  rollinfo_V.nPos=0 ;
	}
	break ;
      case SB_LINEDOWN :
	rollinfo_V.nPos +=rollstep ;
	if( rollinfo_V.nPos > rollinfo_V.nMax ){
	  rollinfo_V.nPos=rollinfo_V.nMax ;
	}
	break ;
      case SB_PAGEUP :
	rollinfo_V.nPos -= rollinfo_V.nPage/2 ;
	if( rollinfo_V.nPos < 0 ){
	  rollinfo_V.nPos=0 ;
	}
	break ;
      case SB_PAGEDOWN :
	rollinfo_V.nPos += rollinfo_V.nPage/2 ;
	if( rollinfo_V.nPos > rollinfo_V.nMax ){
	  rollinfo_V.nPos=rollinfo_V.nMax ;
	}
	break ;
      case SB_THUMBTRACK :
	rollinfo_V.nPos = HIWORD(wp) ;
	break ;
	
      default:
	changed=false ;
	break ;
      }
      if( changed==true ){
	rollinfo_V.fMask  = SIF_POS ;
	SetScrollInfo( hwnd, SB_VERT, &rollinfo_V, true) ;
	//InvalidateRect(hwnd , null , true);
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
      }
      break ;
      
    case WM_LBUTTONDOWN:
      SetCapture( hwnd ) ;
      InvalidateRect(hwnd , null , true);
      SetFocus(hwnd) ;
      break ;
      
    case WM_LBUTTONUP:
      InvalidateRect(hwnd , null , true);
      ReleaseCapture() ;
      break ;
      
    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case PANELCMD.LBTNUP :
	for( int i=0 ; i<gplist.length ; i++ ){
	  if( gplist[i].getHWND()==cast(HWND)(lp) ){
	    curpanel=i ;
	    gplist[i].setSelected(true) ;
	  }else{
	    gplist[i].setSelected(false) ;
	  }
	}
	hairdb.setCurGID(gplist[curpanel].getGid()) ;
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	break ;
      default : break ;
      }
      break;
	
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int getCurGid(){
    return(gplist[curpanel].getGid()) ;
  }

  int setCurGidPanel(int gid){
    foreach( i, gp ; gplist ){
      if( gp.getGid()==gid ){
	curpanel = cast(int)(i) ;
	return(curpanel) ;
      }
    }
    return(-1) ;
  }

  int addGidPanel(int gid){
    if( gplist.length==0 ){
      gplist.length=gplist.length+1 ;
      curpanel=0 ;
    }else{
      gplist.length=gplist.length+1 ;
      for( int i=cast(int)(gplist.length)-1 ; i>curpanel ; i-- ){
	gplist[i]=gplist[i-1] ;
      }
      curpanel++ ;
    }
    
    GidPanel newpanel ;
    newpanel = new GidPanel("GidPanel") ;
    newpanel.setGid(gid) ;
    addObject(newpanel, cast(HMENU)(0), true, false) ;
    newpanel.setOpManager(opman) ;
    gplist[curpanel] = newpanel ;
    sendmsg( WM_SIZE, 0, (winH<<16 | winW) );
    return(0) ;
  }
  
  int delCurGidPanel(){
    if( gplist.length>0 ){
      delObject(gplist[curpanel]) ;
      for( int i=curpanel ; i<gplist.length-1 ; i++ ){
	gplist[i]=gplist[i+1] ;
      }
      gplist.length=gplist.length-1 ;
      if( curpanel==(gplist.length) ) curpanel-- ;
      sendmsg( WM_SIZE, 0, (winH<<16 | winW) );
      return(0) ;
    }
    return(-1) ;
  }

  int init(){
    for( int i=0 ; i<gplist.length ; i++ ){
      delObject(gplist[i]) ;
    }
    delete gplist ;
    gplist.length= 0 ;
    curpanel  =-1 ;
    sendmsg( WM_SIZE, 0, (winH<<16 | winW) );
    return(0) ;
  }

  int syncState(){
    for( int i=0 ; i<gplist.length ; i++ ){
      gplist[i].syncState() ;
    }
    RedrawWindow( getHWND(), null, null, RDW_INVALIDATE) ;
    return(0) ;
  }
    

  int getGidNum(){
    return( cast(int)(gplist.length) ) ;
  }

  int setHairDB(HairDB hairdb){
    if( hairdb is null ) return(-1) ;
    this.hairdb=hairdb ;
    return(0) ;
  }

  int setOpManager(OpManager opman){
    if( opman is null ) return(-1) ;
    this.opman=opman ;
    setHairDB(this.opman.getHairDB()) ;
    return(0) ;
  }

}
