import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

import FontMan ;
import MyButton ;
import GidPanel ;
import GidListPanel ;

import HairDB ;
import HairGuideDB ;
import OpManager ;

COLORREF RGB(uint r, uint g, uint b) {
  return cast(COLORREF)(r | (g << 8) | (b << 16));
}

enum GOP{
  NONE        = 0x100,
  ButtonSEL   = 0x101,
  ButtonUNSEL = 0x102,
  ButtonCHGRP = 0x103
} ;

class GidOpPanel : WindowObject {
  HairDB         hairdb ;
  OpManager      opman ;

  GidListPanel   gidlist ;
  MyButton       btsel   ;
  MyButton       btunsel ;
  MyButton       btchgrp ;

  HFONT          hfont;

  this(string name) {
    winc.style          = 0;
    this.setStyle(WS_CHILD | WS_TABSTOP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)("GidOpPanel") ;
    super(name) ;
    this.registwinc() ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont);
      break ;
      
    case WM_CREATE :
      FontMan font_f12 ; font_f12.init(12,FIXED_PITCH) ;
      hfont = font_f12.createHFONT() ;
      
      btsel= new MyButton("SEL") ;
      btsel.addStyle(BS_CENTER|BS_VCENTER) ;
      btsel.setPosition(0,0,48,18) ;
      addObject(btsel, cast(HMENU)(GOP.ButtonSEL),false,false) ;
      btsel.setFont(hfont) ;
      
      btunsel= new MyButton("UNSEL") ;
      btunsel.addStyle(BS_CENTER|BS_VCENTER) ;
      btunsel.setPosition(48*1,0,48,18) ;
      addObject(btunsel, cast(HMENU)(GOP.ButtonUNSEL),false,false) ;
      btunsel.setFont(hfont) ;

      btchgrp= new MyButton("CHGGRP") ;
      btchgrp.addStyle(BS_CENTER|BS_VCENTER) ;
      btchgrp.setPosition(48*2,0,48,18) ;
      addObject(btchgrp, cast(HMENU)(GOP.ButtonCHGRP),false,false) ;
      btchgrp.setFont(hfont) ;

      gidlist= new GidListPanel("GidListPanel") ;
      gidlist.setPosition(0,32,32,64) ;
      addObject(gidlist, cast(HMENU)(GOP.NONE),false,false) ;
      break ;
      
    case WM_PAINT:
      HDC hdc;
      PAINTSTRUCT ps ;
      HBRUSH BCOLBG ;
      HPEN   PCOLBG ;
      
      BCOLBG = CreateSolidBrush(RGB(64,64,64)) ;
      PCOLBG = CreatePen(PS_SOLID, 1, RGB(64,64,64)) ;
      
      hdc = BeginPaint(hwnd, &ps);
      SelectObject(hdc, PCOLBG) ;
      SelectObject(hdc, BCOLBG) ;
      Rectangle(hdc, 0,0,winW,winH) ;
      EndPaint(hwnd, &ps);

      DeleteObject(BCOLBG) ;
      DeleteObject(PCOLBG) ;
      break ;

    case WM_SIZE:
      MoveWindow(gidlist.getHWND(), 0,18, winW, winH-18, true) ;
      break;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case GOP.ButtonSEL:
	opman.execCommand(OPMCMD.SELECTGID,getCurGid()) ;
	SetFocus(getMainWinHWND()) ;
	break ;
      case GOP.ButtonUNSEL:
	opman.execCommand(OPMCMD.UNSELECTGID,getCurGid()) ;
	SetFocus(getMainWinHWND()) ;
	break ;
      case GOP.ButtonCHGRP:
	GIDSTATE curgidstat = hairdb.getCurGIDstat() ;
	if( curgidstat.lock==false && curgidstat.hide==false ){
	  opman.execCommand(OPMCMD.CHANGEGROUPE,getCurGid()) ;
	}
	SetFocus(getMainWinHWND()) ;
	break ;
      default :	break ;
      }
      break;
	
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int getCurGid(){
    if( gidlist !is null ){
      return(gidlist.getCurGid()) ;
    }else{
      return(-1) ;
    }
  }

  int setCurGidPanel(int gid){
    if( gidlist !is null ){
      return(gidlist.setCurGidPanel(gid)) ;
    }else{
      return(-1) ;
    }
  }

  int addGidPanel(int gid){
    return( gidlist.addGidPanel(gid) ) ;
  }
  
  int delCurGidPanel(){
    return( gidlist.delCurGidPanel() ) ;
  }

  int initGidPanel(){
    return( gidlist.init() ) ;
  }

  int getGidNum(){
    return( gidlist.getGidNum() ) ;
  }

  int syncState(){
    if( gidlist !is null ){
      gidlist.syncState() ;
      return(0) ;
    }
    return(-1) ;
  }

  int resyncHairDB(){
    int current_gidnum=getGidNum() ;
    
    for( int i=0 ; i<current_gidnum ; i++ ){
      delCurGidPanel() ;
    }
    GIDSTATE[] gidstatlist=hairdb.getGIDstatlist() ;
    
    for( int i=0 ; i<gidstatlist.length ; i++ ){
      addGidPanel(i) ;
    }
    setCurGidPanel( hairdb.getCurGID() ) ;
    return(0) ;
  }

  int setHairDB(HairDB hairdb){
    if( hairdb is null ) return(-1) ;
    this.hairdb=hairdb ;
    return(0) ;
  }

  int setOpManager(OpManager opman){
    if( opman is null ) return(-1) ;
    this.opman=opman ;
    setHairDB(this.opman.getHairDB()) ;
    if( gidlist !is null ){
      gidlist.setOpManager(opman) ;
    }
    return(0) ;
  }
}
