import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowMenu;
import win32frame.WindowObject;
import win32frame.PushButton;
import std.stdio;
import std.string;
import std.utf;

import FontMan;
import MyButton ;
import ColorSelectDialog;
import TextEditDialog ;

import HairDB;
import HairGuideDB;
import OpManager ;

COLORREF RGB(uint r, uint g, uint b) {
  return cast(COLORREF)(r | (g << 8) | (b << 16));
}

enum BTNSTAT{
  LOCK = 0x101 ,
  SHOW = 0x102 
} ;

enum PANELCMD{
  LBTNUP  = 0x201 ,
  CHGSTAT = 0x202 ,
  SYNCSTAT= 0x203
} ;

enum POPUP{
  CHGNAME  = 0x301 ,
  CHGCOLOR = 0x302 
} ;

class GidPanel : WindowObject {
  HairDB     hairdb ;
  OpManager  opman ;

  int        mlbtn_state ;
  int        gid ;

  HFONT      btnfont ;
  HFONT      labelfont ;
  bool       selected     =false ;
  bool       mousetracking=false ;
  MyButton   tgl_lock ;
  MyButton   tgl_show ;
  
  this(string name) {
    winc.style          = 0;
    this.setStyle(WS_CHILD | WS_TABSTOP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)("GidPanel") ;
    super(name) ;
    this.registwinc() ;
  }
  
  int setSelected(bool sel){
    selected = sel ;
    RedrawWindow( getHWND(), null, null, RDW_INVALIDATE) ;
    return(0) ;
  }

  int movePanel(int x,int y,int w,int h, bool redraw=true){
    MoveWindow( getHWND(), x, y, w, h, redraw) ;
    return(0) ;
  }
  
  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      delObject(tgl_lock) ;
      delObject(tgl_show) ;
      DeleteObject(btnfont) ;
      DeleteObject(labelfont) ;
      break ;
      
    case WM_CREATE :
      FontMan font_f10 ; font_f10.init(10,FIXED_PITCH) ;
      FontMan font_f12 ; font_f12.init(12,FIXED_PITCH) ;
      btnfont   = font_f10.createHFONT() ;
      labelfont = font_f12.createHFONT() ;

      tgl_lock = new MyButton("FREE") ;
      tgl_lock.addStyle(BS_CENTER|BS_VCENTER) ;
      tgl_lock.setPosition(1,1,13*3,15) ;
      addObject(tgl_lock, cast(HMENU)(BTNSTAT.LOCK)) ;
      tgl_lock.setFont(btnfont) ;
      
      tgl_show = new MyButton("SHOW") ;
      tgl_show.addStyle(BS_CENTER|BS_VCENTER) ;
      tgl_show.setPosition(1,16,13*3,15) ;      
      addObject(tgl_show, cast(HMENU)(BTNSTAT.SHOW)) ;
      tgl_show.setFont(btnfont) ;
      
      break ;
      
    case WM_PAINT:
      HDC    hdc;
      PAINTSTRUCT ps ;
      HBRUSH COLBASE ;
      HBRUSH DISPCOL ;
      COLORREF basecol ;

      GIDSTATE gidstat ;
      GIDDISP  giddisp ;
      if( hairdb !is null ){
	gidstat=hairdb.getGIDstat(gid) ;
	giddisp=hairdb.getGIDdisp(gid) ;
      }
      
      basecol = selected==true ? RGB(255,0,255) : mousetracking==true ? RGB(0,255,255) : GetSysColor(COLOR_3DFACE) ;
      COLBASE = CreateSolidBrush(basecol) ;
      DISPCOL = CreateSolidBrush(giddisp.color) ;
      
      hdc = BeginPaint(hwnd, &ps);
      SelectObject(hdc, COLBASE) ;
      Rectangle(hdc, 0,0,winW,winH) ;
      SelectObject(hdc, DISPCOL) ;
      Rectangle(hdc,42,0,31+42,32) ;
      SetBkMode(hdc , TRANSPARENT);
      {
	SIZE labelsize ;
	wchar[] label=(toUTF16(gidstat.name)).dup ;
//	wchar[] dbgid=(toUTF16(format("gid=%d",gid))).dup ;

	SelectObject(hdc,labelfont) ;
	GetTextExtentPoint32( hdc, cast(wchar*)(label), cast(int)(label.length), &labelsize);
	TextOut(hdc,
		75,0,
		cast(wchar*)(label), cast(int)(label.length)) ;
//	TextOut(hdc,
//		75,16,
//		cast(wchar*)(dbgid), cast(int)(dbgid.length)) ;
      }
      
      EndPaint(hwnd, &ps);
      
      DeleteObject(COLBASE) ;
      DeleteObject(DISPCOL) ;
      break ;

    case WM_SIZE:
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_MOUSELEAVE:
      mousetracking=false ;
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;

    case WM_MOUSEMOVE:
      if( mousetracking==false ){
	TRACKMOUSEEVENT trackmsevent ;
	trackmsevent.cbSize  = TRACKMOUSEEVENT.sizeof ;
	trackmsevent.dwFlags = TME_LEAVE ;
	trackmsevent.hwndTrack = hwnd ;
	TrackMouseEvent( &trackmsevent ) ;
	mousetracking=true ;
      }
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;
      
    case WM_LBUTTONDOWN:
      mlbtn_state=1 ;
      SetCapture( hwnd ) ;
      InvalidateRect(hwnd , null , true);
      SetFocus(hwnd) ;
      break ;
      
    case WM_LBUTTONUP:
      mlbtn_state=0 ;
      InvalidateRect(hwnd , null , true);
      sendmsgtoparent(WM_COMMAND, PANELCMD.LBTNUP, cast(LPARAM)(hwnd)) ;
      ReleaseCapture() ;
      break ;

    case WM_RBUTTONDOWN:
      POINT pos ;
      auto menuroot = new WindowMenu() ;
      menuroot.createPopup() ;
//      menuroot.addItem( MFT_STRING, "Change Name"  ,  POPUP.CHGNAME  ) ;
      menuroot.addItem( MFT_STRING, "Change Color" ,  POPUP.CHGCOLOR)  ;
      
      pos.x = LOWORD(lp);
      pos.y = HIWORD(lp);
      ClientToScreen(hwnd , &pos);
      TrackPopupMenu(menuroot.getHMENU() , TPM_LEFTALIGN | TPM_TOPALIGN ,
		     pos.x , pos.y , 0 , hwnd , null);
      delete menuroot ;
      return(0) ;
      
    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case POPUP.CHGNAME :
	GIDSTATE gidstat = hairdb.getGIDstat(gid) ;
	RECT wrect ;
        GetWindowRect(this.getMainWinHWND(),&wrect) ;
	auto txtdlog = new TextEditDialog("", "txteditdlgbox") ;
	txtdlog.setDefaultText(gidstat.name) ;
	txtdlog.Create( wrect.left+100, wrect.top+100, hwnd, null ) ;
	if( txtdlog.getDlogResult()==1 ){
	  string str = txtdlog.getText() ;
	  if( str.length!=0 ){
	    gidstat.name=str ;
	    hairdb.setGIDstat(gid,gidstat) ;
	  }
	}
	delete txtdlog ;
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE ) ;
	break ;
      case POPUP.CHGCOLOR :
	GIDDISP giddisp = hairdb.getGIDdisp(gid) ;
	RECT wrect ;
        GetWindowRect(this.getMainWinHWND(),&wrect) ;
        auto colsel = new ColorSelectDialog("","colsel") ;
	colsel.setColor(giddisp.color) ;
	colsel.Create( wrect.left+100,wrect.top+100,hwnd,null ) ;
	if( colsel.getDlogResult()==1 ){
          giddisp.color = colsel.getColor() & 0x00ffffff ;
        }
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE ) ;
	hairdb.setGIDdisp(gid,giddisp) ;
	opman.execCommand(OPMCMD.REDRAW) ;
        break ;
      case BTNSTAT.LOCK :
	GIDSTATE gidstat = hairdb.getGIDstat(gid) ;
	if( gidstat.lock==false ){
	  hairdb.selectGid(gid,false) ;
	  gidstat.lock = true ;
	  tgl_lock.setWindowText("LOCK") ;
	}else{
	  gidstat.lock = false ;
	  tgl_lock.setWindowText("FREE") ;
	}
	hairdb.setGIDstat(gid,gidstat) ;
	opman.execCommand(OPMCMD.CHGLOCKSTAT) ;
	SetFocus(getMainWinHWND()) ;
	break;
      case BTNSTAT.SHOW :
	GIDSTATE gidstat = hairdb.getGIDstat(gid) ;
	if( gidstat.hide==false ){
	  hairdb.selectGid(gid,false) ;
	  gidstat.hide = true ;
	  tgl_show.setWindowText("HIDE") ;
	}else{
	  gidstat.hide = false ;
	  tgl_show.setWindowText("SHOW") ;
	}
	hairdb.setGIDstat(gid,gidstat) ;
	opman.execCommand(OPMCMD.CHGSHOWSTAT) ;
	SetFocus(getMainWinHWND()) ;
	break;
      case PANELCMD.SYNCSTAT :
	GIDSTATE gidstat = hairdb.getGIDstat(gid) ;
	if( gidstat.lock==false ){
	  tgl_lock.setWindowText("FREE") ;
	}else{
	  tgl_lock.setWindowText("LOCK") ;
	}
	if( gidstat.hide==false ){
	  tgl_show.setWindowText("SHOW") ;
	}else{
	  tgl_show.setWindowText("HIDE") ;
	}
	RedrawWindow( tgl_lock.getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW ) ;
	RedrawWindow( tgl_show.getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW ) ;
	break ;

      default :	break ;
      }
      break;
	
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int syncState(){
    sendmsg(WM_COMMAND, PANELCMD.SYNCSTAT, 0) ;
    return(0) ;
  }

  int setGid(int gid){
    this.gid = gid ;
    return(0) ;
  }

  int getGid(){
    return(this.gid) ;
  }

  int setHairDB(HairDB hairdb){
    if( hairdb is null ) return(-1) ;
    this.hairdb=hairdb ;
    return(0) ;
  }

  int setOpManager(OpManager opman){
    if( opman is null ) return(-1) ;
    this.opman=opman ;
    setHairDB(this.opman.getHairDB()) ;
    return(0) ;
  }

}
