import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.utf;

import Vec3D;
import Camera;
public import SearchST;

public import HairGuideDB;
public import HairRootDB;

import FileHair;
import FileOBJ ;
public import FilePOV ;

enum EDMD{
  NONE= 0,
  HR  = 1,
  HG  = 2
} ;

class HairDB{
  int         maxgid ;
  HairRootDB  hrDB ;
  HairGuideDB hgDB ;
  int         cur_gid  ;
  EDMD        cur_edmd ;
  HGEDMD      cur_hgedmd ;
  string      errmsg ;
  SearchST    searchID ;
  Camera      camera ;
  string      fname ;
  POVParam    povparam ;

  this(int maxgid){
    hrDB = new HairRootDB() ;
    hgDB = new HairGuideDB(maxgid) ;
    this.maxgid = maxgid ;
    cur_gid=0 ;
    cur_edmd  =EDMD.HR ;
    cur_hgedmd=HGEDMD.LINEAR ;
    fname    ="NewFile.hair" ;
  }

  int setCurGID(int gid){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    cur_gid=gid ;
    return(0) ;
  }

  int getCurGID(){
    return(cur_gid) ;
  }

  GIDSTATE[] getGIDstatlist(){
    return(hgDB.getGIDstatlist()) ;
  }

  GIDSTATE getGIDstat(int gid){
    return(hgDB.getGIDstat(gid)) ;
  }

  int setGIDstat(int gid, GIDSTATE gidstate){
    return(hgDB.setGIDstat(gid, gidstate)) ;
  }

  GIDDISP[] getGIDdisplist(){
    return(hgDB.getGIDdisplist()) ;
  }

  GIDDISP getGIDdisp(int gid){
    return(hgDB.getGIDdisp(gid)) ;
  }

  int setGIDdisp(int gid, GIDDISP giddisp){
    return(hgDB.setGIDdisp(gid, giddisp)) ;
  }

  GIDSTATE getCurGIDstat(){
    return(hgDB.getGIDstat(cur_gid)) ;
  }

  int setEditMode(EDMD mode){
    cur_edmd=mode ;
    return(0) ;
  }

  EDMD getEditMode(){
    return(cur_edmd) ;
  }

  int setHGEditMode(HGEDMD mode){
    cur_hgedmd=mode ;
    return(0) ;
  }

  HGEDMD getHGEditMode(){
    return(cur_hgedmd) ;
  }

  int changeGid(int gid){
    int ret=hgDB.changeGid(gid) ;
    return(ret) ;
  }
    
  string getCurFilename(){
    return(fname) ;
  }
    
  int selectAll(bool sel=true){
    int ret ;
    switch( cur_edmd ){
    case EDMD.HR :
      ret=hrDB.selectAll(sel) ;
      break ;
    case EDMD.HG :
      ret=hgDB.selectAll(sel) ;
      break ;
    default : core.exception.onSwitchError() ;
    }
    return(ret) ;
  }

  int selectGid(int gid, bool sel=true){
    return(hgDB.selectGid(gid,sel)) ;
  }

  int select(SearchST id, bool sel=true){
    int ret ;
    switch( cur_edmd ){
    case EDMD.HR :
      ret=hrDB.select(id,sel) ;
      break ;
    case EDMD.HG :
      ret=hgDB.select(id,sel) ;
      break ;
    default : core.exception.onSwitchError() ;
    }
    return(ret) ;
  }

  int checkSelected(SearchST id){
    int ret ;
    switch( cur_edmd ){
    case EDMD.HR :
      ret=hrDB.checkSelected(id) ;
      break ;
    case EDMD.HG :
      ret=hgDB.checkSelected(id) ;
      break ;
    default : core.exception.onSwitchError() ;
    }
    return(ret) ;
  }

  int setDrawHR(vec3D eye, vec3D ray){
    return( hrDB.setDrawHR(eye,ray) ) ;
  }

  SearchST search(vec3D eye, vec3D ray){
    SearchST ret ;
    switch( cur_edmd ){
    case EDMD.HR :
      ret=hrDB.search(eye,ray) ;
      break ;
    case EDMD.HG :
      ret=hgDB.search(eye,ray) ;
      break ;
    default : core.exception.onSwitchError() ;
    }
    searchID=ret ;
    return(ret) ;
  }

  SearchST getSearchID(){
    return(searchID) ;
  }

  int move(SearchST HGid, vec3D eye, vec3D frray, vec3D toray){
    int ret=0 ;
    switch( cur_edmd ){
    case EDMD.HG :
      ret=hgDB.move(HGid,eye,frray,toray,cur_hgedmd) ;
      break ;
    default : break ;
    }
    return(ret) ;
  }
  
  int scale(double scale, HGEDMD hgedmd){
    int ret=0 ;
    switch( cur_edmd ){
    case EDMD.HG :
      ret=hgDB.scale(scale,hgedmd) ;
      break ;
    default : break ;
    }
    return(ret) ;
  }
  
  int generateHG(){
    int gencount=0 ;
    if( cur_edmd==EDMD.HR ){
      vec3D[] HRvtx ; HRvtx = hrDB.getHRvtx() ;
      vec3D[] HRnml ; HRnml = hrDB.getHRnml() ;
      for( int i=0 ; i<HRvtx.length ; i++ ){
	if( HRvtx[i].sel==true ){
	  int ret=hgDB.generate( HRvtx[i], HRnml[i], HRvtx[i].id, cur_gid ) ;
	  if( ret==0 ) gencount++ ;
	}
      }
    }
    return(gencount) ;
  }

  int deleteHG(){
    int ret=0 ;
    if( cur_edmd==EDMD.HG ){
      ret=hgDB.del() ;
    }
    return(ret) ;
  }

  int genSurfaceList(){
    return(hgDB.genSurfaceList()) ;
  }

  int delSurfaceList(){
    return(hgDB.delSurfaceList()) ;
  }

  int[] getSurfaceList(){
    return(hgDB.getSurfaceList()) ;
  }

  vec3D[] getHRvtx(){
    return( hrDB.getHRvtx() ) ;
  }
  
  face3D[] getHRface(){
    return( hrDB.getHRface() ) ;
  }

  vec3D[] getHRfnml(){
    return( hrDB.getHRfnml() ) ;
  }

  HairGuide[] getHGlist(){
    return( hgDB.getHGlist() ) ;
  }

  HairGuide[] getFloatHGlist(){
    return( hgDB.getFloatHGlist() ) ;
  }

  void setHideHRface(bool hide){
    hrDB.setHideHRface(hide) ;
  }
  bool getHideHRface(){
    return(hrDB.getHideHRface()) ;
  }
  
  void setHideHRvtx(bool hide){
    hrDB.setHideHRvtx(hide) ;
  }
  bool getHideHRvtx(){
    return( hrDB.getHideHRvtx() ) ;
  }
  
  void setHideHGline(bool hide){
    hgDB.setHideHGline(hide) ;
  }
  bool getHideHGline(){
    return( hgDB.getHideHGline() ) ;
  }
  
  void setHideHGvtx(bool hide){
    hgDB.setHideHGvtx(hide) ;
  }
  bool getHideHGvtx(){
    return( hgDB.getHideHGvtx() ) ;
  }

  int read_hair(string fname){
    FileHair filehair = new FileHair() ;
    try{
      filehair.read(fname,maxgid) ;
      hrDB = filehair.getHairRootDB()  ;
      hgDB = filehair.getHairGuideDB() ;
    }catch(Exception e){
      errmsg=e.msg ;
      return(-1) ;
    }
    this.fname=fname ;
    return(0) ;
  }

  int read_obj(string fname){
    FileOBJ fileobj = new FileOBJ() ;
    try{
      fileobj.read(fname,maxgid) ;
      hrDB = fileobj.getHairRootDB()  ;
      hgDB = fileobj.getHairGuideDB() ;
    }catch(Exception e){
      errmsg=e.msg ;
      return(-1) ;
    }
    return(0) ;
  }

  int write_hair(string fname){
    FileHair filehair = new FileHair() ;
    try{
      filehair.write(fname, hrDB, hgDB) ;
    }catch(Exception e){
      errmsg=e.msg ;
      return(-1) ;
    }
    this.fname=fname ;
    return(0) ;
  }

  int write_pov(string fname, POVParam param){
    FilePOV filepov = new FilePOV() ;
    try{
      filepov.write(fname, hrDB, hgDB, param) ;
      povparam=param ;
    }catch(Exception e){
      errmsg=e.msg ;
      return(-1) ;
    }
    return(0) ;
  }

  bool isObjExists(){
    return( hrDB.isObjExists()==true || hgDB.isObjExists()==true ) ;
  }

  HairGuideDB cloneHairGuideDB(){
    return(hgDB.cloneDB()) ;
  }

  HairRootDB cloneHairRootDB(){
    return(hrDB.cloneDB()) ;
  }

  POVParam getPOVparam(){
    return(povparam) ;
  }

  string getErrorMsg(){
    return(errmsg) ;
  }

  int setCamera(Camera camera){
    if( camera is null ) return(-1) ;
    this.camera=camera ;
    return(0) ;
  }
}
