import std.stdio;
import std.string;
import std.utf;
import std.math;

import Vec3D;
import SearchST;

uint RGB(uint r, uint g, uint b) {
  return cast(uint)(r | (g << 8) | (b << 16));
}

struct HairGuide{
  int       type  ;
  int       gid   ;
  bool      sel   ;
  int       HRid  ;
  vec3D[]   plist ;
  double[]  knots ;
} ;

class GIDDISP {
  float  r=1.0 ;
  float  g=1.0 ;
  float  b=1.0 ;
  float  a=1.0 ;
  uint   color=0x00ffffff ;
  this(){} ;
} ;

class GIDSTATE{
  bool   hide=false ;
  bool   lock=false ;
  string name ;

  this(){} ;

  GIDSTATE clone(){
    auto clonegidstat = new GIDSTATE() ;
    clonegidstat.hide=this.hide ;
    clonegidstat.lock=this.lock ;
    clonegidstat.name=this.name ;
    return(clonegidstat) ;
  }
} ;

struct SelectHistory{
  int[]  selhis  ;
  int    selptr  ;
  int    selnum  ;
  int[]  suflist ;

  void init(){
    selhis.length=32 ;
    selptr=0 ;
    selnum=0 ;
  }
  void add(int hgid){
    bool found=false ;
    for( int i=0 ; i<selnum ; i++ ){
      if( selhis[(selptr-(i+1)+selhis.length)%selhis.length]==hgid ){
	found=true ;
	break ;
      }
    }
    if( found==false ){
      selhis[selptr]=hgid ;
      selptr=(selptr+1)%cast(int)(selhis.length) ;
      selnum = selnum>=selhis.length ? selnum : selnum+1 ;
    }
  }
  void reset(){
    selptr=0 ;
    selnum=0 ;
  }
  int delsuf(){
    int count=cast(int)(suflist.length) ;
    delete suflist ;
    suflist.length=0 ;
    return(count) ;
  }
  int gensuf(){
    delete suflist ;
    suflist.length=selnum ;
    for( int i=0  ; i<selnum ; i++ ){
      suflist[i]=selhis[(selptr-(i+1)+selhis.length)%selhis.length] ;
    }
    return(cast(int)(suflist.length)) ;
  }

  SelectHistory clone(){
    SelectHistory clonehis = this ;
    clonehis.selhis   = new int[selhis.length] ;
    clonehis.selhis[] = selhis[] ;
    clonehis.suflist  = new int[suflist.length] ;
    clonehis.suflist[]= suflist[] ;
    return(clonehis) ;
  }

} ;

enum HGEDMD{
  LINEAR  =0,
  SYMMETRY,
  MIRROR,
  SCALE,   //this is dummy.
  SCALEINI,
  SCALING,
  SCALEFIX
} ;

double CHOP(double a){return(floor(a*100000000000000.0)/100000000000000.0);}
const int    POINT_NUM = (3*2+1) ;

class HairGuideDB{
  private HairGuide[]   HGlist ;
  private HairGuide[]   floatHGlist ;
  private int           maxgid ;
  private GIDSTATE[]    gid_statlist ;
  private GIDDISP[]     gid_displist ;
  private bool          hideHGline=false;
  private bool          hideHGvtx =false;
  private vec3D[]       HRnml ;
  private SelectHistory selhis ;
  private bool          lockdb=false ;

  const double GEN_STEP     = 0.1  ;
  const double SEARCH_RADIUS= 0.05 ;

  this(int maxgid){
    HGlist.length=0 ;
    floatHGlist.length=0 ;
    if( maxgid<=0 ){
      this.maxgid = 1 ;
    }else{
      this.maxgid = maxgid ;
    }
    
    gid_displist.length = this.maxgid ;
    foreach( i, ref gid_disp ; gid_displist ){
      gid_disp = new GIDDISP() ;
      uint   color ;
      switch(i%4){
      case 0:  color=RGB(255,255,  0) ; break ;
      case 1:  color=RGB(255,  0,255) ; break ;
      case 2:  color=RGB(  0,255,  0) ; break ;
      case 3:  color=RGB(  0,255,255) ; break ;
      default : core.exception.onSwitchError() ;
      }
      gid_disp.color=color ;
    }

    gid_statlist.length = this.maxgid ;
    foreach( i, ref gid_stat ; gid_statlist ){
      gid_stat = new GIDSTATE() ;
      gid_stat.name =format("Group%d",i) ;
    }
    
    selhis.init() ;
  }

  ~this(){
    for( int i=0 ; i<HGlist.length ; i++ ){
      delete HGlist[i].plist ;
      delete HGlist[i].knots ;
    }
    delete HGlist ;
    delete gid_statlist ;
  }

  int generate( vec3D hrvtx, vec3D hrnml, int id, int gid ){
    if( gid_statlist[gid].lock==true || gid_statlist[gid].hide==true ) return(-1) ;
    
    vec3D pstep ; pstep = hrvtx ;
    vec3D nstep ; nstep = hrnml.normalize() * GEN_STEP ;
    vec3D[] plist = new vec3D[POINT_NUM] ;
    foreach( ref p ; plist ){
      p = pstep ;
      pstep+=(nstep) ;
    }
    int rc= add( id, gid, true, plist ) ;
    return(rc) ;
  }
  
  int add( int id, int gid, bool sel, vec3D[] plist ){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    HGlist.length     = HGlist.length+1 ;
    HGlist[$-1].type  = 0    ;
    HGlist[$-1].gid   = gid  ;
    HGlist[$-1].sel   = sel  ;
    HGlist[$-1].HRid  = id   ;
    HGlist[$-1].knots.length=8 ;
    for( int i=0 ; i<HGlist[$-1].knots.length ; i++ ){
      if( i<(HGlist[$-1].knots.length/2) ){
	HGlist[$-1].knots[i]=0.0 ;
      }else{
	HGlist[$-1].knots[i]=1.0 ;
      }
    }
    HGlist[$-1].plist = plist ;
    return(0) ;
  }

  int del(){
    if( lockdb==true ) return(0) ;
    int j=0 ;
    int delcount=0 ;
    for( int i=0 ; i<HGlist.length ; i++ ){
      if( HGlist[i].sel==false ){
	HGlist[j]=HGlist[i] ;
	j++ ;
      }else{
	delete HGlist[i].plist ;
	delete HGlist[i].knots ;
	delcount++ ;
      }
    }
    HGlist.length=j ;
    selhis.reset() ;
    return(delcount) ;
  }

  int changeGid(int gid){
    int chgcount=0 ;
    if( gid<0 || gid>=maxgid ) return(-1) ;
    foreach( ref hg ; HGlist ){
      if( hg.sel==true ){
	if( hg.gid!=gid ) chgcount++ ;
	hg.gid=gid ;
      }
    }
    return(chgcount) ;
  }

  int selectAll(bool sel=true){
    int count=0 ;
    foreach( ref hg ; HGlist ){
      if( gid_statlist[hg.gid].lock==false && gid_statlist[hg.gid].hide==false ){
	if( hg.sel!=sel ) count++ ;
	hg.sel=sel ;
      }
    }
    if( sel==false ) selhis.reset() ;
    return(count) ;
  }
  
  int selectGid(int gid, bool sel=true){
    int count=0 ;
    if( gid<0 || gid>=maxgid ) return(-1) ;
    foreach( ref hg ; HGlist ){
      if( hg.gid==gid && gid_statlist[gid].lock==false && gid_statlist[hg.gid].hide==false ){
	if( hg.sel!=sel ) count++ ;
	hg.sel=sel ;
      }
    }
    return(count) ;
  }

  int select(SearchST HGid, bool sel=true){
    int count=0 ;
    if( HGid.id0<0 || HGid.id0>=HGlist.length ) return(-1) ;
    if( gid_statlist[HGlist[HGid.id0].gid].lock==false && gid_statlist[HGlist[HGid.id0].gid].hide==false ){
      if( HGlist[HGid.id0].sel!=sel ) count++ ;
      HGlist[HGid.id0].sel=sel ;
      selhis.add(HGid.id0) ;
    }
    return(count) ;
  }

  int checkSelected(SearchST HGid){
    if( HGid.id0<0 || HGid.id0>=HGlist.length ) return(-1) ;
    if( gid_statlist[HGlist[HGid.id0].gid].lock==false && gid_statlist[HGlist[HGid.id0].gid].hide==false ){
      if( HGlist[HGid.id0].sel==true ) return(1) ;
    }
    return(0) ;
  }

  int move(SearchST HGid, vec3D eye, vec3D frray, vec3D toray, HGEDMD mode){
    int rc=0 ;
    if( HGid.id0!=-1 ){
      switch(mode){
      case HGEDMD.LINEAR :
	rc=moveLinear(HGid, eye, frray, toray) ;
	break ;
      case HGEDMD.SYMMETRY :
	rc=moveSymmetry(HGid, eye, frray, toray) ;
	break ;
      case HGEDMD.MIRROR :
	rc=moveMirror(HGid, eye, frray, toray) ;
	break ;
      default : core.exception.onSwitchError() ;
      }
    }
    return(rc) ;
  }

  int scale(double scale, HGEDMD mode){
    int rc=0 ;
    switch(mode){
    case HGEDMD.SCALEINI :
      rc=scaleInit() ;
      break ;
    case HGEDMD.SCALING :
      rc=scaleMove(scale) ;
      break ;
    case HGEDMD.SCALEFIX :
      rc=scaleFix() ;
      break ;
    default : core.exception.onSwitchError() ;
    }
    return(rc) ;
  }

  private int moveLinear(SearchST HGid, vec3D eye, vec3D frray, vec3D toray){
    vec3D fvec = HGlist[HGid.id0].plist[HGid.id1] - eye ;
    vec3D mvec = toray * (fvec.length()/frray.length()) - fvec   ;
    int   chgcount=0 ;
    
    foreach( i, ref hg ; HGlist ){
      if( i==HGid.id0 || hg.sel==true ){
	chgcount++ ;
	switch(HGid.id1){
	case 0:
	  break ;
	case 2:
	  hg.plist[HGid.id1  ]+=mvec ;
	  hg.plist[HGid.id1+2]-=mvec ;
	  break ;
	case 3:
	  hg.plist[HGid.id1-1]+=mvec ;
	  hg.plist[HGid.id1  ]+=mvec ;
	  hg.plist[HGid.id1+1]+=mvec ;
	  break ;
	case 4:
	  hg.plist[HGid.id1-2]-=mvec ;
	  hg.plist[HGid.id1  ]+=mvec ;
	  break ;
	case 1:
	case 5:
	  hg.plist[HGid.id1  ]+=mvec ;
	  break ;
	case 6:
	  hg.plist[HGid.id1-1]+=mvec ;
	  hg.plist[HGid.id1  ]+=mvec ;
	  break ;
	default : break ;
	}
      }
    }
    return(chgcount) ;
  }

  private int moveSymmetry(SearchST HGid, vec3D eye, vec3D frray, vec3D toray){
    vec3D fvec = HGlist[HGid.id0].plist[HGid.id1] - eye ;
    vec3D mvec = toray * (fvec.length()/frray.length()) - fvec   ;
    int   chgcount=0 ;
    
    chgcount++ ;
    switch(HGid.id1){
    case 0:
      break ;
    case 2:
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1+2]-=mvec ;
      break ;
    case 3:
      HGlist[HGid.id0].plist[HGid.id1-1]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1+1]+=mvec ;
      break ;
    case 4:
      HGlist[HGid.id0].plist[HGid.id1-2]-=mvec ;
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      break ;
    case 1:
    case 5:
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      break ;
    case 6:
      HGlist[HGid.id0].plist[HGid.id1-1]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      break ;
    default : break ;
    }

    vec3D  sprjXZ    = HRnml[HGlist[HGid.id0].HRid] ; sprjXZ.y=0.0 ; sprjXZ=sprjXZ.normalize() ;
    vec3D  srot_toXZ = HRnml[HGlist[HGid.id0].HRid] ;
    double scosth    = CHOP(srot_toXZ.dot(sprjXZ)) ;
    scosth = scosth<-1.0 ? -1.0 : (scosth>1.0 ? 1.0 : scosth) ;
    double sth_toXZ  = acos( scosth ) ;
    srot_toXZ=(srot_toXZ.cross(sprjXZ)).normalize() ;
    
    foreach( i, ref hg ; HGlist ){
      if( i!=HGid.id0 && hg.sel==true ){
	chgcount++ ;
	vec3D  dprjXZ    =HRnml[hg.HRid] ; dprjXZ.y=0.0  ; dprjXZ=dprjXZ.normalize() ;
	vec3D  drot_toXZ =HRnml[hg.HRid] ;
	double dcosth    = CHOP(drot_toXZ.dot(dprjXZ)) ;
	dcosth = dcosth<-1.0 ? -1.0 : (dcosth>1.0 ? 1.0 : dcosth) ;
	double dth_toXZ  = acos( dcosth ) ;
	drot_toXZ=(drot_toXZ.cross(dprjXZ)).normalize() ;

	double cosyth=CHOP(sprjXZ.dot(dprjXZ)) ;
	cosyth = cosyth<-1.0 ? -1.0 : (cosyth>1.0 ? 1.0 : cosyth) ;
	double yth   = acos(cosyth) ;
	vec3D  yaxis = (sprjXZ.cross(dprjXZ)).normalize() ;

	vec3D p0 = hg.plist[0] ;
	for( int j=0 ; j<hg.plist.length ; j++ ){
	  hg.plist[j]=HGlist[HGid.id0].plist[j] - HGlist[HGid.id0].plist[0] ;
	  if( sth_toXZ!=0.0 ) hg.plist[j]=hg.plist[j].rotate( sth_toXZ, srot_toXZ ) ;
	  if( yth     !=0.0 ){
	    if( cosyth==-1.0 ){
	      yaxis.set(0.0,1.0,0.0) ;
	      hg.plist[j]=hg.plist[j].rotate( yth     , yaxis     ) ;
	    }else{
	      hg.plist[j]=hg.plist[j].rotate( yth     , yaxis     ) ;
	    }
	  }
	  if( dth_toXZ!=0.0 ) hg.plist[j]=hg.plist[j].rotate(-dth_toXZ, drot_toXZ ) ;
	  hg.plist[j]+=p0 ;

	}
      }
    }
    return(chgcount) ;
  }

  private int moveMirror(SearchST HGid, vec3D eye, vec3D frray, vec3D toray){
    vec3D fvec = HGlist[HGid.id0].plist[HGid.id1] - eye ;
    vec3D mvec = toray * (fvec.length()/frray.length()) - fvec   ;
    int   chgcount=0 ;
    
    chgcount++ ;
    switch(HGid.id1){
    case 0:
      break ;
    case 2:
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1+2]-=mvec ;
      break ;
    case 3:
      HGlist[HGid.id0].plist[HGid.id1-1]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1+1]+=mvec ;
      break ;
    case 4:
      HGlist[HGid.id0].plist[HGid.id1-2]-=mvec ;
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      break ;
    case 1:
    case 5:
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      break ;
    case 6:
      HGlist[HGid.id0].plist[HGid.id1-1]+=mvec ;
      HGlist[HGid.id0].plist[HGid.id1  ]+=mvec ;
      break ;
    default : break ;
    }

    foreach( i, ref hg ; HGlist ){
      if( i!=HGid.id0 && hg.sel==true ){
	chgcount++ ;
	vec3D   master_HRnml = HRnml[HGlist[HGid.id0].HRid] ;
	vec3D   master_p0    = HGlist[HGid.id0].plist[0] ;
	vec3D[] master_plist = new vec3D[POINT_NUM] ;
	master_plist[] = HGlist[HGid.id0].plist[] ;

	vec3D p0 = hg.plist[0] ;
	if( !((master_p0.x<0.0 && p0.x<0.0) || (master_p0.x>=0.0 && p0.x>=0.0)) ){
	  master_HRnml.x = -master_HRnml.x ;
	  foreach( ref p ; master_plist ){
	    p.x = -p.x ;
	  }
	}

	vec3D  sprjXZ    = master_HRnml ; sprjXZ.y=0.0 ; sprjXZ=sprjXZ.normalize() ;
	vec3D  srot_toXZ = master_HRnml ;
	double scosth    = CHOP(srot_toXZ.dot(sprjXZ)) ;
	scosth = scosth<-1.0 ? -1.0 : (scosth>1.0 ? 1.0 : scosth) ;
	double sth_toXZ  = acos( scosth ) ;
	srot_toXZ=(srot_toXZ.cross(sprjXZ)).normalize() ;
    
    	vec3D  dprjXZ    = HRnml[hg.HRid] ; dprjXZ.y=0.0  ; dprjXZ=dprjXZ.normalize() ;
	vec3D  drot_toXZ = HRnml[hg.HRid] ;
	double dcosth    = CHOP(drot_toXZ.dot(dprjXZ)) ;
	dcosth = dcosth<-1.0 ? -1.0 : (dcosth>1.0 ? 1.0 : dcosth) ;
	double dth_toXZ  = acos( dcosth ) ;
	drot_toXZ=(drot_toXZ.cross(dprjXZ)).normalize() ;

	double cosyth=CHOP(sprjXZ.dot(dprjXZ)) ;
	cosyth = cosyth<-1.0 ? -1.0 : (cosyth>1.0 ? 1.0 : cosyth) ;
	double yth   = acos(cosyth) ;
	vec3D  yaxis = (sprjXZ.cross(dprjXZ)).normalize() ;

	for( int j=0 ; j<hg.plist.length ; j++ ){
	  hg.plist[j]=(master_plist[j] - master_plist[0]) ;
	  if( sth_toXZ!=0.0 ) hg.plist[j]=hg.plist[j].rotate( sth_toXZ, srot_toXZ ) ;
	  if( yth     !=0.0 ){
	    if( cosyth==-1.0 ){
	      yaxis.set(0.0,1.0,0.0) ;
	      hg.plist[j]=hg.plist[j].rotate( yth     , yaxis     ) ;
	    }else{
	      hg.plist[j]=hg.plist[j].rotate( yth     , yaxis     ) ;
	    }
	  }
	  if( dth_toXZ!=0.0 ) hg.plist[j]=hg.plist[j].rotate(-dth_toXZ, drot_toXZ ) ;
	  hg.plist[j]+=p0 ;

	}
      }
    }
    return(chgcount) ;
  }

  private int scaleInit(){
    lockdb=true ;
    floatHGlist = new HairGuide[HGlist.length] ;
    foreach( i, hg ; HGlist ){
      floatHGlist[i]=hg ;
      floatHGlist[i].plist  = new vec3D[POINT_NUM] ;
      floatHGlist[i].plist[]=hg.plist[] ;
      floatHGlist[i].knots  = new double[8] ;
      floatHGlist[i].knots[]=hg.knots[] ;
    }
    return(0) ;
  }

  private int scaleMove(double scale){
    int chgcount=0 ;

    if( scale!=1.0 ){
      foreach( i, ref hg ; floatHGlist ){
	if( hg.sel==true ){
	  for( int j=1 ; j<hg.plist.length ; j++ ){
	    hg.plist[j] = ((HGlist[i].plist[j]-HGlist[i].plist[0])*scale)+HGlist[i].plist[0] ;
	  }
	  chgcount++ ;
	}
      }
    }
    return(chgcount) ;
  }

  private int scaleFix(){
    for( int i=0 ; i<HGlist.length ; i++ ){
      HGlist[i]=floatHGlist[i] ;
    }
    delete floatHGlist ;
    floatHGlist.length=0 ;
    lockdb=false ;
    return(0) ;
  }

  SearchST search(vec3D eye, vec3D ray){
    double   mindst ;
    SearchST searched ;
    foreach( i, hg ; HGlist ){
      if( gid_statlist[hg.gid].hide ==true ||
	  gid_statlist[hg.gid].lock==true ) continue ;
      for( int j=0 ; j<hg.plist.length ; j++ ){
	vec3D vp = hg.plist[j] - eye ;
	double qleng = vp.dot(ray)/ray.length ;
	if( qleng<0.0 ) continue ;
	double dist=(ray.normalize() * qleng - vp).length() ;
	if( dist<SEARCH_RADIUS ){
	  if( searched.id0==-1 ){
	    mindst=dist ;
	    searched.id0=cast(int)(i) ;
	    searched.id1=j ;
	  }else{
	    if( mindst>dist ){
	      mindst=dist ;
	      searched.id0=cast(int)(i) ;
	      searched.id1=j ;
	    }
	  }
	}
      }
    }
    return(searched) ;
  }

  bool isObjExists(){
    bool result=false ;
    if( HGlist.length!=0 ){
      result=true ;
    }
    return(result) ;
  }

  HairGuideDB cloneDB(){
    auto clonedb = new HairGuideDB(this.maxgid) ;
    clonedb.HGlist = new HairGuide[HGlist.length] ;
    foreach( i, hg ; HGlist ){
      clonedb.HGlist[i]=hg ;
      clonedb.HGlist[i].plist   = new vec3D[POINT_NUM] ;
      clonedb.HGlist[i].plist[] = hg.plist[] ;
      clonedb.HGlist[i].knots   = new double[8] ;
      clonedb.HGlist[i].knots[] = hg.knots[] ;
    }

    foreach( i, gid_stat ; gid_statlist ){
      clonedb.gid_statlist[i] = gid_stat.clone() ;
    }
    clonedb.gid_displist[] = gid_displist[] ;
    clonedb.hideHGline = hideHGline ;
    clonedb.hideHGvtx  = hideHGvtx ;
    clonedb.HRnml      = new vec3D[HRnml.length] ;
    clonedb.HRnml[]    = HRnml[] ;
    clonedb.selhis     = selhis.clone() ;
    clonedb.lockdb     = false ;
    return(clonedb) ;
  }

  int setGIDcolor(int gid, uint color){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    gid_displist[gid].color=color ;
    return(0) ;
  }

  uint getGIDcolor(int gid){
    if( gid<0 || gid>=maxgid ) return(0) ;
    return(gid_displist[gid].color) ;
  }
  
  int setGIDhide(int gid, bool hide){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    gid_statlist[gid].hide=hide ;
    return(0) ;
  }
  
  int setGIDlock(int gid, bool lock){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    gid_statlist[gid].lock=lock ;
    return(0) ;
  }

  int setGIDname(int gid, string name){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    gid_statlist[gid].name=name ;
    return(0) ;
  }

  string getGIDname(int gid){
    if( gid<0 || gid>=maxgid ) return("") ;
    return(gid_statlist[gid].name) ;
  }

  int getGIDnum(){
    return(maxgid) ;
  }

  GIDSTATE[] getGIDstatlist(){
    return(gid_statlist) ;
  }

  GIDSTATE getGIDstat(int gid){
    return(gid_statlist[gid]) ;
  }

  int setGIDstat(int gid, GIDSTATE gidstate){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    gid_statlist[gid]=gidstate ;
    return(0) ;
  }

  GIDDISP[] getGIDdisplist(){
    return(gid_displist) ;
  }

  GIDDISP getGIDdisp(int gid){
    return(gid_displist[gid]) ;
  }

  int setGIDdisp(int gid, GIDDISP giddisp){
    if( gid<0 || gid>=maxgid ) return(-1) ;
    gid_displist[gid]=giddisp ;
    return(0) ;
  }

  HairGuide[] getHGlist(){
    return(HGlist) ;
  }
  HairGuide[] getFloatHGlist(){
    return(floatHGlist) ;
  }
  
  void setHideHGline(bool hide){
    hideHGline=hide ;
  }
  bool getHideHGline(){
    return(hideHGline) ;
  }
  
  void setHideHGvtx(bool hide){
    hideHGvtx=hide ;
  }
  bool getHideHGvtx(){
    return(hideHGvtx) ;
  }

  void setHRnmlList(vec3D[] hrnml){
    HRnml=hrnml ;
  }

  int genSurfaceList(){
    return(selhis.gensuf()) ;
  }
  
  int delSurfaceList(){
    return(selhis.delsuf()) ;
  }
  
  int[] getSurfaceList(){
    return(selhis.suflist) ;
  }

}
