import std.stdio;
import std.string;
import std.utf;

import Vec3D;
import SearchST;

class HairRootDB{
  private vec3D[]   HRvtx  ;
  private vec3D[]   HRnml  ;
  private face3D[]  HRface ;
  private vec3D[]   HRfnml ;
  private bool      hideHRface=false;
  private bool      hideHRvtx =false;

  const double SEARCH_RADIUS=0.05 ;
  
  this(){
  }

  ~this(){
    delete HRvtx  ;
    delete HRnml  ;
    delete HRface ;
    delete HRfnml ;
  }

  int set( vec3D[] HRvtx, face3D[] HRface ){
    this.HRvtx = HRvtx ;
    foreach( i, ref v ; this.HRvtx ){
      v.id =cast(int)(i) ;
      v.sel=false ;
    }
    this.HRface = HRface ;
    HRfnml.length=HRface.length ;
    HRnml.length = HRvtx.length ;
    foreach( ref n ; HRnml ){
	n.set(0.0,0.0,0.0) ;
    }
    foreach( i, f ; HRface ){
      vec3D vec = ((HRvtx[f.p2] - HRvtx[f.p1]).cross(HRvtx[f.p1] - HRvtx[f.p0])).normalize() ;
      HRfnml[i]=vec ;
      HRnml[f.p0]+=vec ;
      HRnml[f.p1]+=vec ;
      HRnml[f.p2]+=vec ;
    }
    foreach( ref n ; HRnml ){
      if( n.length()==0.0 ){
	n.set(1.0,0.0,0.0) ;
      }else{
	n=n.normalize() ;
      }
    }
    return(0) ;
  }

  int selectAll(bool sel=true){
    int count=0 ;
    for( int i=0 ; i<HRvtx.length ; i++ ){
      if( HRvtx[i].sel!=sel ) count++ ;
      HRvtx[i].sel=sel ;
    }
    return(count) ;
  }

  int select(SearchST HRid, bool sel=true){
    int count=0 ;
    if( HRid.id0<0 || HRid.id0>=HRvtx.length ) return(-1) ;
    if( HRvtx[HRid.id0].sel!=sel ) count++ ;
    HRvtx[HRid.id0].sel=sel ;
    return(count) ;
  }

  int checkSelected(SearchST HRid){
    if( HRid.id0<0 || HRid.id0>=HRvtx.length ) return(-1) ;
    if( HRvtx[HRid.id0].sel==true ) return(1) ;
    return(0) ;
  }

  int setDrawHR(vec3D eye, vec3D ray){
    foreach( ref v ; HRvtx ){
      v.disp=false ;
    }
    foreach( i, fn ; HRfnml ){
      double res = ray.dot(fn) ;
      if( res<=0.0 ){
	HRvtx[HRface[i].p0].disp=true ;
	HRvtx[HRface[i].p1].disp=true ;
	HRvtx[HRface[i].p2].disp=true ;
      }
    }
    return(0) ;
  }
  
  SearchST search(vec3D eye, vec3D ray){
    SearchST searched ;
    double   mindst ;
    int      fid=-1 ;
    //視線と交差するfaceの中から最も視点に近いfaceを探す
    foreach( i, f ; HRface ){
      vec3D crosstmp=ray ;
      if( crosstmp.dot(HRfnml[i])<=0.0 ){
	if( crosstmp.LPcross( eye, ray, HRvtx[f.p0], HRfnml[i] )==true ){
	  if( crosstmp.chkinside( HRvtx[f.p0], HRvtx[f.p1], HRvtx[f.p2] )==true ){
	    crosstmp-=eye ;
	    if( fid==-1 ){
	      mindst=crosstmp.length() ;
	      fid=cast(int)(i) ;
	    }else{
	      if( mindst>crosstmp.length() ){
		mindst=crosstmp.length() ;
		fid=cast(int)(i) ;
	      }
	    }
	  }
	}
      }
    }
    if( fid!=-1 ){
      vec3D[] nearvtxlist = new vec3D[3] ;
      nearvtxlist[0]=HRvtx[HRface[fid].p0] ;
      nearvtxlist[1]=HRvtx[HRface[fid].p1] ;
      nearvtxlist[2]=HRvtx[HRface[fid].p2] ;
    
      int vid=-1 ;
      //探した視点に最も近いfaceの3頂点の中からSEARCH_RADIUS内に入っている頂点
      //で最も視線との距離が近い頂点を探す
      foreach( i, nearvtx ; nearvtxlist ){
	//if( nearvtx[i].disp==false ) continue ;
	vec3D  vp    = nearvtx - eye ;
	double qleng = vp.dot(ray)/ray.length() ;
	if( qleng<0.0 ) continue ;
	double dist=(ray.normalize() * qleng - vp).length() ;
	if( dist<SEARCH_RADIUS ){
	  if( vid==-1 ){
	    mindst=dist ;
	    vid=cast(int)(i) ;
	  }else{
	    if( mindst>dist ){
	      mindst = dist ;
	      vid=cast(int)(i) ;
	    }
	  }
	}
      }
      if( vid!=-1 ){
	searched.id1=fid  ;
	switch(vid){
	case 0: searched.id0=HRface[fid].p0 ; break ;
	case 1: searched.id0=HRface[fid].p1 ; break ;
	case 2: searched.id0=HRface[fid].p2 ; break ;
	default : break ;
	}
	return(searched) ;
      }
    }
    return(searched) ;
  }

  bool isObjExists(){
    bool result=false ;
    if( HRvtx.length!=0 || HRface.length!=0 ){
      result=true ;
    }
    return(result) ;
  }

  HairRootDB cloneDB(){
    auto clonedb = new HairRootDB() ;
    clonedb.HRvtx   = new vec3D[HRvtx.length] ;
    clonedb.HRvtx[] = HRvtx[] ;
    clonedb.HRnml   = new vec3D[HRnml.length] ;
    clonedb.HRnml[] = HRnml[] ;
    clonedb.HRface  = new face3D[HRface.length] ;
    clonedb.HRface[]= HRface[] ;
    clonedb.HRfnml  = new vec3D[HRfnml.length] ;
    clonedb.HRfnml[]= HRfnml[] ;
    clonedb.hideHRface = hideHRface ;
    clonedb.hideHRvtx  = hideHRvtx ;
    return(clonedb) ;
  }
  
  vec3D[] getHRvtx(){
    return( HRvtx ) ;
  }
  
  vec3D[] getHRnml(){
    return( HRnml ) ;
  }
  
  face3D[] getHRface(){
    return( HRface ) ;
  }
  
  vec3D[] getHRfnml(){
    return( HRfnml ) ;
  }

  void setHideHRface(bool hide){
    hideHRface=hide ;
  }
  bool getHideHRface(){
    return(hideHRface) ;
  }
  
  void setHideHRvtx(bool hide){
    hideHRvtx=hide ;
  }
  bool getHideHRvtx(){
    return(hideHRvtx) ;
  }
  
}
