/*
 *
 */
import std.stdio;
import std.string;
import std.math;
import Vec3D;
import Material;

class Object3D
{
  vec3D[]     vtx   ;
  vec3D[]     nml   ;
  vec2D[]     uvm   ;
  face3D[][]  faceV ;
  face3D[][]  faceN ;
  face3D[][]  faceT ;
  vec3D[]     fn    ;
  Material[]  mat   ;
  int[char[]] matNameToID ;
  
  int   vtx_num  ;
  int   nml_num  ;
  int   uvm_num  ;
  int[] face_num ;
  int   matID    ;

  double scale   ;
  
  this(){
    this.scale=1.0 ;
    vtx_num =0 ;
    nml_num =0 ;
    uvm_num =0 ;
    matID   =0 ;
  }

  this(double scale){
    this.scale=scale ;
    vtx_num =0 ;
    nml_num =0 ;
    uvm_num =0 ;
    matID   =0 ;
  }

  ~this(){
    delete vtx   ;
    delete nml   ;
    delete uvm   ;
    delete faceV ;
    delete faceN ;
    delete faceT ;
    delete fn    ;
    delete mat   ;
    delete face_num ;
  }
  
  int init(){
    return(0) ;
  }
  
  int add_vertex(double x, double y, double z){
    if( vtx.length == 0 ){
      vtx.length=256 ;
    }else{
      if( vtx_num >= vtx.length ){
	vtx.length = vtx.length * 2 ;
      }
    }
    vtx[vtx_num].x =x*scale ;
    vtx[vtx_num].y =y*scale ;
    vtx[vtx_num].z =z*scale ;
    vtx[vtx_num].id=vtx_num ;
    vtx_num++ ;
    return(vtx_num) ;
  }
  
  int add_texuv(double u, double v){
    if( uvm.length == 0 ){
      uvm.length=256 ;
    }else{
      if( uvm_num >= uvm.length ){
	uvm.length = uvm.length * 2 ;
      }
    }
    uvm[uvm_num].u=    u ;
    uvm[uvm_num].v=1.0-v ;
    uvm_num++ ;
    return(uvm_num) ;
  }

      
  int add_normal(double x, double y, double z){
    if( nml.length == 0 ){
      nml.length=256 ;
    }else{
      if( nml_num >= nml.length ){
	nml.length = nml.length * 2 ;
      }
    }
    nml[nml_num].x = x ;
    nml[nml_num].y = y ;
    nml[nml_num].z = z ;
    nml_num++ ;
    return(nml_num) ;
  }

  int add_material(string matname){
    int* idp ;
    Color clr ;

    idp = (matname in matNameToID) ;
    if( idp == null ){
      matNameToID[matname]=matID++ ;
    }
    
    clr.r=1.0;clr.g=1.0;clr.b=1.0;clr.a=1.0;
    
    mat.length=mat.length+1 ;
    mat[matID-1].Kd = clr ;
    mat[matID-1].Ka = clr ;
    mat[matID-1].Ks = clr ;
    
    mat[matID-1].tex.length=0 ;
    return(0) ;
  }

  int add_mtl_Ns(double ns){
    mat[matID-1].Ns=ns ;
    return(0) ;
  }
  
  int add_mtl_Kd(double r, double g, double b){
    mat[matID-1].Kd.r=r ;
    mat[matID-1].Kd.g=g ;
    mat[matID-1].Kd.b=b ;
    return(0) ;
  }
  
  int add_mtl_Ka(double r, double g, double b){
    mat[matID-1].Ka.r=r ;
    mat[matID-1].Ka.g=g ;
    mat[matID-1].Ka.b=b ;
    return(0) ;
  }
  
  int add_mtl_Ks(double r, double g, double b){
    mat[matID-1].Ks.r=r ;
    mat[matID-1].Ks.g=g ;
    mat[matID-1].Ks.b=b ;
    return(0) ;
  }
  
  int add_mtl_mapKd(string mapfile){
    mat[matID-1].tex=mapfile ;
    return(0) ;
  }
  
  int finish_material(){
    faceV.length    = matID ;
    faceN.length    = matID ;
    faceT.length    = matID ;
    face_num.length = matID ;
    
    return(0) ;
  }
  
  int add_face(string matname, face3D f, face3D t, face3D n){
    int* idp ;
    int  mtlid ;

    idp = (matname in matNameToID) ;
    if( idp == null ){
      writef("material %s is not found\n", matname) ;
      return(-1) ;
    }
    mtlid = *idp ;
    
    if( faceV[mtlid].length==0 ){
      faceV[mtlid].length=256 ;
      faceT[mtlid].length=256 ;
      faceN[mtlid].length=256 ;
    }else{
      if( face_num[mtlid] >= faceV[mtlid].length ){
	faceV[mtlid].length = faceV[mtlid].length * 2 ;
	faceT[mtlid].length = faceT[mtlid].length * 2 ;
	faceN[mtlid].length = faceN[mtlid].length * 2 ;
      }
    }
    faceV[mtlid][face_num[mtlid]].p0 = f.p0-1 ;
    faceV[mtlid][face_num[mtlid]].p1 = f.p1-1 ;
    faceV[mtlid][face_num[mtlid]].p2 = f.p2-1 ;
    faceT[mtlid][face_num[mtlid]].p0 = t.p0-1 ;
    faceT[mtlid][face_num[mtlid]].p1 = t.p1-1 ;
    faceT[mtlid][face_num[mtlid]].p2 = t.p2-1 ;
    faceN[mtlid][face_num[mtlid]].p0 = n.p0-1 ;
    faceN[mtlid][face_num[mtlid]].p1 = n.p1-1 ;
    faceN[mtlid][face_num[mtlid]].p2 = n.p2-1 ;
    face_num[mtlid]++ ;
    return(face_num[mtlid]) ;
  }
  
  int finish(){
    int i ;
    vec3D tmp1,tmp2 ;
    
    vtx.length = vtx_num ;
    nml.length = nml_num ;
    uvm.length = uvm_num ;

    foreach( mtlid ; matNameToID ){
      faceV[mtlid].length=face_num[mtlid] ;
      faceN[mtlid].length=face_num[mtlid] ;
      faceT[mtlid].length=face_num[mtlid] ;
    }
    return(0) ;
  }

  void printInfo(){
    writef("  vtx = %d\n",vtx.length) ;
    writef("  nml = %d\n",nml.length) ;
    writef("  uvm = %d\n",uvm.length) ;
    foreach( mtlname ; matNameToID.keys ){
      writef("  face[%s]=%d\n",mtlname,faceV[ matNameToID[mtlname] ].length) ;
    }
  }
}
