import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.glut;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import Vec3D;
import Camera;
import myMainWindow;
import OpenGLFrame;
import OpPanel;
import HairDB;
import WorkSpace;
import UndoBuffer;

enum HMOP {
  SEARCH   = 0 ,
  SELECTINI,
  SELECTING,
  SELECTFIX,
  UNSELECTINI,
  UNSELECTING,
  UNSELECTFIX,
  MOVEINI,
  MOVING,
  MOVEFIX,
  SCALEINI,
  SCALING,
  SCALEFIX,
  REDRAW
};

enum OPMCMD {
  VIEWRESET = 0x200 ,
  SELALL,
  UNSELALL,
  GENHG,
  TGL_HIDE_OBJ,
  TGL_HIDE_HGCP,
  TGL_EDITMODE,
  CHGEDITMODE,
  TGL_HGEDMODE,
  CHGHGEDMODE,
  GENSURFACE,
  DELSURFACE,
  DELETE,
  SET_HIDE_HR,
  SET_HIDE_HGCP,
  SET_HIDE_OBJ,
  SELECTGID,
  UNSELECTGID,
  CHANGEGROUPE,
  CHGLOCKSTAT,
  CHGSHOWSTAT,
  UNDO,
  REDO,
  REDRAW,
  EVENTSHORTCUT
} ;

class OpManager {
  HairDB       hairdb ;
  OpenGLFrame  oglframe ;
  OpPanel      oppanel ;
  myMainWindow mainwin ;
  UndoBuffer   undobuf ;
  bool         db_changed ;
  
  this() {
  }

  int checkSelected(int x, int y, int prev_x, int prev_y ){
    SearchST searchID ;
    Camera   camera=hairdb.camera ;
    camera.calc(x, y, prev_x, prev_y) ;

    searchID=hairdb.search(camera.geyepos, camera.toray) ;
    return(hairdb.checkSelected(searchID)) ;
  } 
  
  int updateDB( HMOP mode, int x, int y, int prev_x, int prev_y ){
    SearchST searchID ;
    Camera   camera=hairdb.camera ;
    camera.calc(x, y, prev_x, prev_y) ;
    hairdb.setDrawHR(camera.geyepos,camera.geyeray) ;
    double   scale = 1.0+ cast(double)(x-prev_x)/100.0 ;

    switch( mode ){
    case HMOP.SELECTINI :
    case HMOP.UNSELECTINI :
      db_changed=false ;
      break ;
    case HMOP.SELECTING :
      searchID=hairdb.search(camera.geyepos, camera.toray) ;
      int ret=hairdb.select(searchID,true) ;
      if( ret>0 ) db_changed=true ;
      break ;
    case HMOP.UNSELECTING :
      searchID=hairdb.search(camera.geyepos, camera.toray) ;
      int ret=hairdb.select(searchID,false) ;
      if( ret>0 ) db_changed=true ;
      break ;
    case HMOP.SELECTFIX :
    case HMOP.UNSELECTFIX :
      if( db_changed==true ){
	undobuf.push(new UndoPack(hairdb.cloneHairRootDB(),hairdb.cloneHairGuideDB()),false) ;
      }
      break ;

    case HMOP.MOVEINI :
      db_changed=false ;
      break ;
    case HMOP.MOVING :
      searchID = hairdb.getSearchID() ;
      int ret = hairdb.move(searchID,camera.geyepos, camera.frray, camera.toray) ;
      if( ret>0 ){
	oppanel.printStatusbar(format("Edit %d hairguide(s)",ret)) ;
	db_changed=true ;
      }
      break ;
    case HMOP.MOVEFIX :
      if( db_changed==true ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),true) ;
      }
      oppanel.printStatusbar("") ;
      break ;

    case HMOP.SEARCH :
    case HMOP.REDRAW :
      searchID=hairdb.search(camera.geyepos, camera.toray) ;
      break ;

    case HMOP.SCALEINI :
      hairdb.scale(scale,HGEDMD.SCALEINI) ;
      db_changed=false ;
      break ;
    case HMOP.SCALING :
      int ret=hairdb.scale(scale,HGEDMD.SCALING) ;
      if( ret>0 ){
	oppanel.printStatusbar(format("Scale %.0f%% %d hairguides(s)",scale*100.0,ret)) ;
	db_changed=true ;
      }else{
	db_changed=false ;
      }
      break ;
    case HMOP.SCALEFIX :
      hairdb.scale(scale,HGEDMD.SCALEFIX) ;
      if( db_changed==true ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),true) ;
      }
      oppanel.printStatusbar("") ;
      break ;

    default : core.exception.onSwitchError() ;
    }
    mainwin.setWindowTitle() ;
    oppanel.sendmsg(WM_COMMAND, OPPCMD.NONE, 0) ; //update UNDO/REDO button state
    return(1) ;
  }

  int keyboard( int type, int acs, int key, int x, int y ){
    if( type==0 ){
      switch(key){
      case 'r' : execCommand(OPMCMD.VIEWRESET     ) ; break ;
      case 'a' : execCommand(OPMCMD.SELALL        ) ; break ;
      case 'g' : execCommand(OPMCMD.GENHG         ) ; break ;
      case 'h' : execCommand(OPMCMD.TGL_HIDE_OBJ  ) ; break ;
      case 'H' : execCommand(OPMCMD.TGL_HIDE_HGCP ) ; break ;
      case 'y' : execCommand(OPMCMD.GENSURFACE    ) ; break ;
      case 'x' : execCommand(OPMCMD.DELSURFACE    ) ; break ;
      default :
	break ;
      }
    }else{
      if(       (acs & CNTL_KEY)!=0 && (acs & SHIFT_KEY)!=0 ){
	if( key=='Z') execCommand(OPMCMD.REDO) ;
      }else if( (acs & CNTL_KEY)!=0 && (acs & SHIFT_KEY)==0 ){
	if( key=='Z') execCommand(OPMCMD.UNDO) ;
      }else{
	switch(key){
	case VK_SPACE  : execCommand(OPMCMD.UNSELALL ) ; break ;
	case VK_DELETE : execCommand(OPMCMD.DELETE   ) ; break ;
	default :
	  break ;
	}
      }
    }
    return(1) ;
  }

  int execCommand(OPMCMD cmd, int param0=0, int param1=0){
    switch(cmd){
    case OPMCMD.VIEWRESET :
      hairdb.camera.reset() ;
      updateDB( HMOP.REDRAW, 0, 0, 0, 0 ) ;
      break ;
    case OPMCMD.SELALL : 
      int rc=hairdb.selectAll(true) ;
      if( rc>0 ){
	undobuf.push(new UndoPack(hairdb.cloneHairRootDB(),hairdb.cloneHairGuideDB()),false) ;
      }
      break ;
    case OPMCMD.UNSELALL : 
      int rc=hairdb.selectAll(false) ;
      if( rc>0 ){
	undobuf.push(new UndoPack(hairdb.cloneHairRootDB(),hairdb.cloneHairGuideDB()),false) ;
      }
      break ;
    case OPMCMD.GENHG :
      EDMD cur_edmode = getEditMode() ;
      setEditMode(EDMD.HR) ;
      int rc=hairdb.generateHG() ;
      if( rc>0 ){
	hairdb.selectAll(false) ;
	oppanel.printStatusbar(format("Generate %d hairguide(s)",rc)) ;
	undobuf.push(new UndoPack(hairdb.cloneHairRootDB(),hairdb.cloneHairGuideDB()),true) ;
      }
      setEditMode(cur_edmode) ;
      break ;
    case OPMCMD.TGL_HIDE_OBJ :
      hairdb.setHideHRface( (hairdb.getHideHRface()==false ? true : false) ) ;
      break ;
    case OPMCMD.TGL_HIDE_HGCP :
      hairdb.setHideHGvtx( (hairdb.getHideHGvtx()==false ? true : false) ) ;
      break ;
    case OPMCMD.TGL_EDITMODE :
      switch(hairdb.getEditMode()){
      case EDMD.HR : hairdb.setEditMode(EDMD.HG) ; break ;
      case EDMD.HG : hairdb.setEditMode(EDMD.HR) ; break ;
      default : core.exception.onSwitchError() ;
      }
      break ;
    case OPMCMD.TGL_HGEDMODE :
      switch(hairdb.getHGEditMode()){
      case HGEDMD.LINEAR   : hairdb.setHGEditMode(HGEDMD.SYMMETRY) ; break ;
      case HGEDMD.SYMMETRY : hairdb.setHGEditMode(HGEDMD.LINEAR  ) ; break ;
      default : core.exception.onSwitchError() ;
      }
      break ;
    case OPMCMD.GENSURFACE :
      int rc=hairdb.genSurfaceList() ;
      if( rc>0 ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),false) ;
      }
      break ;
    case OPMCMD.DELSURFACE :
      int rc=hairdb.delSurfaceList() ;
      if( rc>0 ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),false) ;
      }
      break ;
    case OPMCMD.DELETE :
      EDMD cur_edmode = getEditMode() ;
      setEditMode(EDMD.HG) ;
      int rc=hairdb.deleteHG() ;
      if( rc>0 ){
	hairdb.delSurfaceList() ;
	oppanel.printStatusbar(format("Delete %d hairguide(s)",rc)) ;
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),true) ;
      }
      setEditMode(cur_edmode) ;
      break ;

    case OPMCMD.SET_HIDE_HR :
      hairdb.setHideHRvtx(param0==0 ? false : true) ;
      break ;
    case OPMCMD.SET_HIDE_HGCP :
      hairdb.setHideHGvtx(param0==0 ? false : true) ;
      break ;
    case OPMCMD.SET_HIDE_OBJ :
      hairdb.setHideHRface(param0==0 ? false : true) ;
      break ;

    case OPMCMD.SELECTGID :
      if( hairdb.selectGid(param0,true)>0 ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),false) ;
      }
      break ;
    case OPMCMD.UNSELECTGID :
      if( hairdb.selectGid(param0,false)>0 ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),false) ;
      }
      break ;
    case OPMCMD.CHANGEGROUPE :
      if( hairdb.changeGid(param0)>0 ){
	undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),true) ;
      }
      break ;
    case OPMCMD.CHGLOCKSTAT :
      undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),false) ;
      updateDB( HMOP.REDRAW, 0, 0, 0, 0 ) ;
      break ;
    case OPMCMD.CHGSHOWSTAT :
      undobuf.push(new UndoPack(null,hairdb.cloneHairGuideDB()),false) ;
      updateDB( HMOP.REDRAW, 0, 0, 0, 0 ) ;
      break ;
      
    case OPMCMD.UNDO :
      auto undodt=cast(UndoPack)(undobuf.undo()) ;
      if( undodt !is null ){
	if( undodt.hrDB !is null ) hairdb.hrDB=undodt.hrDB.cloneDB() ;
	if( undodt.hgDB !is null ) hairdb.hgDB=undodt.hgDB.cloneDB() ;
	oppanel.syncState() ;
	updateDB( HMOP.REDRAW, 0, 0, 0, 0 ) ;
	oppanel.printStatusbar("Undo") ;
      }else{
	oppanel.printStatusbar("Nothing to Undo") ;
      }
      break ;
      
    case OPMCMD.REDO :
      auto redodt=cast(UndoPack)(undobuf.redo()) ;
      if( redodt !is null ){
	if( redodt.hrDB !is null ) hairdb.hrDB=redodt.hrDB.cloneDB() ;
	if( redodt.hgDB !is null ) hairdb.hgDB=redodt.hgDB.cloneDB() ;
	oppanel.syncState() ;
	updateDB( HMOP.REDRAW, 0, 0, 0, 0 ) ;
	oppanel.printStatusbar("Redo") ;
      }else{
	oppanel.printStatusbar("Nothing to Redo") ;
      }
      break ;
      
    case OPMCMD.REDRAW:
      updateDB( HMOP.REDRAW, 0, 0, 0, 0 ) ;
      break ;
    default :
      return(0) ;
      break ;
    }
    oglframe.redraw() ;
    mainwin.setWindowTitle() ;
    return(0) ;
  }

  int setEditMode(EDMD edmd){
    return(hairdb.setEditMode(edmd) ) ;
  }

  EDMD getEditMode(){
    return(hairdb.getEditMode() ) ;
  }

  int setHGEditMode(HGEDMD hgedmd){
    return( hairdb.setHGEditMode(hgedmd) ) ;
  }

  int setCntlKeyLocked(bool locked){
    if( oglframe is null ) return(-1) ;
    return( oglframe.setCntlKeyLocked(locked) ) ;
  }

  int setScaleMode(bool mode){
    if( oglframe is null ) return(-1) ;
    return( oglframe.setScaleMode(mode) ) ;
  }

  int setHairDB(HairDB hairdb){
    if( hairdb is null ) return(-1) ;
    this.hairdb=hairdb ;
    return(0) ;
  }

  HairDB getHairDB(){
    return(hairdb) ;
  }

  int setUndoBuffer(UndoBuffer undobuf){
    if( undobuf is null ) return(-1) ;
    this.undobuf=undobuf ;
    return(0) ;
  }

  int setOpenGLFrame(OpenGLFrame oglframe){
    if( oglframe is null ) return(-1) ;
    this.oglframe=oglframe ;
    return(0) ;
  }
    
  int setOpPanel(OpPanel oppanel){
    if( oppanel is null ) return(-1) ;
    this.oppanel=oppanel ;
    return(0) ;
  }
    
  int setMainWindow(myMainWindow mainwin){
    if( mainwin is null ) return(-1) ;
    this.mainwin=mainwin ;
    return(0) ;
  }
    
}
