import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import win32frame.ToolTip;
import win32frame.Launcher;
import win32frame.Split;
import win32frame.Panel;
import util.Image;
import util.BitMap ;
import std.stdio;
import std.string;
import std.math;
import std.utf;

import FontMan ;
import MyButton ;
import MyStaticCtl ;
import TextEditDialog ;
import ColorSelectDialog ;

import OpManager ;
import HairDB ;
import GidOpPanel ;
import UndoBuffer;

enum OPPCMD{
  NONE    = 0x0,
  SELHR,
  GENHG,
  SELHG,
  MOVLINE,
  MOVSYM,
  MOVMIRROR,
  MOVSCALE,
  DELHG,
  GENSUF,
  DELSUF,
  UNDO,
  REDO
} ;

struct ToolButton{
  int           type  ;
  MyButton      btn   ;
  MyStaticCtl   stc   ;
  ToolTip       tip   ;
  string        resname ;
  string        label ;
  string        tiptxt;
  int           cmd   ;
  
  void set(int type, string resname, string label, string tiptxt, int cmd){
    this.type   =type ;
    this.resname=resname ;
    this.label  =label ;
    this.tiptxt =tiptxt ;
    this.cmd    =cmd ;
  }
} ;

const ALT_KEY       = (0x0001) ;
const CNTL_KEY      = (0x0002) ;
const SHIFT_KEY     = (0x0004) ;

class OpPanel : WindowObject {
  Split         split ;
  Panel         btnpanel ;
  Launcher      oplauncher ;
  ToolButton[]  tblist ;
  GidOpPanel    gidoppanel ;
  
  OpManager     opman ;
  UndoBuffer    undobuf ;

  int           ACS_stat=0 ;
  
  HFONT         hfont_v14,hfont_f12;
  
  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = null ;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont_v14);
      DeleteObject(hfont_f12);
      break ;
      
    case WM_CREATE:
      FontMan font_v14 ; font_v14.init(14,VARIABLE_PITCH) ;
      FontMan font_f12 ; font_f12.init(12,FIXED_PITCH) ;

      hfont_v14 = font_v14.createHFONT() ;
      hfont_f12 = font_f12.createHFONT() ;

      split = new Split("Split",1) ;
      split.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      split.setPosition(0,0,300,200) ;
      split.setSashPos(120,120,55250) ;
      addObject( split, cast(HMENU)(OPPCMD.NONE), true, true) ;

      btnpanel = new Panel("BtnPanel") ;
      btnpanel.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      btnpanel.setPosition(0,0,300,200) ;
      split.addObject( btnpanel, cast(HMENU)(OPPCMD.NONE)) ;      
      
      int[] collist = [ 1,1,1,1 ] ;
      int[] rowlist = [ 1,1,1   ] ;
      oplauncher = new Launcher("opLauncher") ;
      oplauncher.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      oplauncher.setPosition(0,20,300,32*3) ;
      oplauncher.initGrid(collist,rowlist,true) ;
      btnpanel.addObject(oplauncher, cast(HMENU)(OPPCMD.NONE),true,false) ;

      tblist.length=collist.length * rowlist.length ;
      tblist[ 0].set( 1, "TB_SELHR"   , "SelHR"      , "Select HairRoot"       , OPPCMD.SELHR     );
      tblist[ 1].set( 0, "TB_GENHG"   , "Generate"   , "Genarate HairGuide"    , OPPCMD.GENHG     );
      tblist[ 2].set( 1 ,"TB_SELHG"   , "SelHG"      , "Select HairGuide"      , OPPCMD.SELHG     );
      tblist[ 3].set( 0 ,"TB_DELHG"   , "delete"     , "Delete HairGuide"      , OPPCMD.DELHG     );
      tblist[ 4].set( 1 ,"TB_MOVLINE" , "moveLiner"  , "Move Liner"            , OPPCMD.MOVLINE   );
      tblist[ 5].set( 1 ,"TB_MOVSCL"  , "moveScale"  , "Move Scale"            , OPPCMD.MOVSCALE  );
      tblist[ 6].set( 1 ,"TB_MOVMIR"  , "moveMirror" , "Move Mirror"           , OPPCMD.MOVMIRROR );
      tblist[ 7].set( 1 ,"TB_MOVSYM"  , "moveSymtry" , "Move Y-axial symmetry" , OPPCMD.MOVSYM    );
      tblist[ 8].set( 0 ,"TB_GENSUF"  , "genSurface" , "Generate Surface"      , OPPCMD.GENSUF    );
      tblist[ 9].set( 0 ,"TB_DELSUF"  , "delSurface" , "Delete Surface"        , OPPCMD.DELSUF    );
      tblist[10].set( 0 ,"TB_UNDO"    , "Undo"       , "Undo"                  , OPPCMD.UNDO      );
      tblist[11].set( 0 ,"TB_REDO"    , "Redo"       , "Redo"                  , OPPCMD.REDO      );
      //tblist[10].set( 9 ,""           , ""           , ""                      , OPPCMD.NONE      );
      //tblist[11].set( 9 ,""           , ""           , ""                      , OPPCMD.NONE      );

      foreach( i, ref tb ; tblist ){
	if( tb.type==0 || tb.type==1 ){
	  tb.btn = new MyButton(tb.label) ;
	  if( tb.resname != "" ){
	    auto bmp = new BitMap() ;
	    bmp.readImage(getHInstance(), toUTFz!(wchar*)(tb.resname), MAKEINTRESOURCE_T!(256)) ;
	    tb.btn.setBitMap(bmp) ;
	  }
	  tb.btn.addStyle(BS_CENTER|BS_VCENTER) ;
	  tb.btn.setPosition(0,0,32,32) ;
	  //btn.setFlatMode(true) ;
	  oplauncher.addObject(tb.btn, cast(HMENU)(tb.cmd)) ;
	  tb.btn.setFont(hfont_f12) ;
	  tb.tip = new ToolTip() ;
	  tb.tip.addStyle(TTS_BALLOON) ;
	  tb.tip.setText(tb.tiptxt) ;
	  tb.btn.addToolTip(tb.tip) ;
	  tb.tip.setClientRect() ;
	}else{
	  tb.stc = new MyStaticCtl(tb.label) ;
	  tb.stc.addStyle(BS_RIGHT|BS_VCENTER) ;
	  tb.stc.setPosition(0,0,32,32) ;
	  oplauncher.addObject(tb.stc, cast(HMENU)(tb.cmd)) ;
	  tb.stc.setFont(hfont_v14) ;
	}
      }

      gidoppanel = new GidOpPanel("GidOpPanel") ;
      gidoppanel.setPosition(0,280,100,180) ;
      split.addObject(gidoppanel, cast(HMENU)(OPPCMD.NONE)) ;

      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC         hdc;
      hdc = BeginPaint(hwnd, &ps);
      EndPaint(hwnd, &ps);
      break ;
      
    case WM_CHAR:
      if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
      else                            ACS_stat &= ~(SHIFT_KEY) ;
      if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
      else                            ACS_stat &= ~(CNTL_KEY)  ;
      if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
      else                            ACS_stat &= ~(ALT_KEY)   ;
      opman.keyboard( 0, ACS_stat, LOBYTE(cast(ushort)(wp)), 0,0) ;
      break ;

    case WM_KEYDOWN:
      if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
      else                            ACS_stat &= ~(SHIFT_KEY) ;
      if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
      else                            ACS_stat &= ~(CNTL_KEY)  ;
      if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
      else                            ACS_stat &= ~(ALT_KEY)   ;
      opman.keyboard( 1, ACS_stat, LOBYTE(cast(ushort)(wp)), 0,0) ;
      break ;

    case WM_COMMAND:
      int cmd=LOWORD(wp) ;
      switch(cmd){
      case OPPCMD.SELHR:
      case OPPCMD.SELHG:
      case OPPCMD.MOVLINE:
      case OPPCMD.MOVSCALE:
      case OPPCMD.MOVMIRROR:
      case OPPCMD.MOVSYM:
	resetTBstate(cmd) ;
	switch(cmd){
	case OPPCMD.SELHR:
	  opman.setEditMode(EDMD.HR) ;
	  opman.setCntlKeyLocked(true) ;
	  opman.setScaleMode(false) ;
	  break ;
	case OPPCMD.SELHG:
	  opman.setEditMode(EDMD.HG) ;
	  opman.setCntlKeyLocked(true) ;
	  opman.setScaleMode(false) ;
	  break ;
	case OPPCMD.MOVLINE:
	  opman.setEditMode(EDMD.HG) ;
	  opman.setHGEditMode(HGEDMD.LINEAR) ;
	  opman.setCntlKeyLocked(false) ;
	  opman.setScaleMode(false) ;
	  break ;
	case OPPCMD.MOVSCALE:
	  opman.setEditMode(EDMD.HG) ;
	  opman.setHGEditMode(HGEDMD.SCALE) ;
	  opman.setCntlKeyLocked(false) ;
	  opman.setScaleMode(true) ;
	  break ;
	case OPPCMD.MOVMIRROR:
	  opman.setEditMode(EDMD.HG) ;
	  opman.setHGEditMode(HGEDMD.MIRROR) ;
	  opman.setCntlKeyLocked(false) ;
	  opman.setScaleMode(false) ;
	  break ;
	case OPPCMD.MOVSYM:
	  opman.setEditMode(EDMD.HG) ;
	  opman.setHGEditMode(HGEDMD.SYMMETRY) ;
	  opman.setCntlKeyLocked(false) ;
	  opman.setScaleMode(false) ;
	  break ;
	default : break ;
	}
	break ;
      case OPPCMD.GENHG:
	EDMD cur_edmode = opman.getEditMode() ;
	opman.setEditMode(EDMD.HR) ;
	opman.execCommand(OPMCMD.GENHG) ;
	opman.execCommand(OPMCMD.UNSELALL) ;
	opman.setEditMode(cur_edmode) ;
	break ;
      case OPPCMD.DELHG:
	EDMD cur_edmode = opman.getEditMode() ;
	opman.setEditMode(EDMD.HG) ;
	opman.execCommand(OPMCMD.DELETE) ;
	opman.setEditMode(cur_edmode) ;
	break ;
      case OPPCMD.GENSUF:
	opman.execCommand(OPMCMD.GENSURFACE) ;
	break ;
      case OPPCMD.DELSUF:
	opman.execCommand(OPMCMD.DELSURFACE) ;
	break ;
      case OPPCMD.UNDO:
	opman.execCommand(OPMCMD.UNDO) ;
	break ;
      case OPPCMD.REDO:
	opman.execCommand(OPMCMD.REDO) ;
	break ;
      default :
	break ;
      }
      break;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }

    if( undobuf !is null ){
      UNDOSTATE undostat=undobuf.getUndoState() ;
      if( undostat==UNDOSTATE.TOP ){
	tblist[10].btn.enable() ;
	tblist[11].btn.disable() ;
      }else if( undostat==UNDOSTATE.BOTTOM ){
	tblist[10].btn.disable() ;
	tblist[11].btn.enable() ;
      }else if( undostat==UNDOSTATE.NONE ){
	tblist[10].btn.disable() ;
	tblist[11].btn.disable() ;
      }else{
	tblist[10].btn.enable() ;
	tblist[11].btn.enable() ;
      }	
    }

    SetFocus(getMainWinHWND()) ;
    return(0);
  }

  void resetTBstate(int cmd){
    foreach( i, ref tb ; tblist ){
      if( tb.type==1 ){
	tb.btn.setSelectState(false) ;
	if(  tb.cmd==cmd ){
	  tb.btn.setSelectState(true) ;
	}
      }
    }
  }

  int setOpManager(OpManager opman){
    if( opman is null ) return(-1) ;
    this.opman=opman ;
    return(0) ;
  }

  int setUndoBuffer(UndoBuffer undobuf){
    if( undobuf is null ) return(-1) ;
    this.undobuf=undobuf ;
    return(0) ;
  }

  LRESULT syncButtonState(){
    if( opman is null ) return(-1) ;
    return(sendmsg(WM_COMMAND, OPPCMD.SELHR, 0)) ;
  }

  int syncState(){
    if( gidoppanel !is null ){
      gidoppanel.syncState() ;
      return(0) ;
    }
    return(-1) ;
  }
}
