import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0010) ;
const RBUTTONUP     = (0x0020) ;
const RBUTTONDBLCLK = (0x0040) ;
const ALT_KEY       = (0x0100) ;
const CNTL_KEY      = (0x0200) ;
const SHIFT_KEY     = (0x0400) ;
const ACS_KEY_FILTER= (0x0700) ;

class OpenGLBase : WindowObject {
  HGLRC hRC;
  HDC   WinDC;
  PIXELFORMATDESCRIPTOR pfd = { 
    PIXELFORMATDESCRIPTOR.sizeof,    // size of this pfd 
    1,                               // version number 
    PFD_DRAW_TO_WINDOW |             // support window 
    PFD_SUPPORT_OPENGL |             // support OpenGL 
    PFD_DOUBLEBUFFER   |             // double buffered 
    PFD_TYPE_RGBA,                   // RGBA type 
    24,                              // 24-bit color depth 
    0, 0, 0, 0, 0, 0,                // color bits ignored 
    0,                               // no alpha buffer 
    0,                               // shift bit ignored 
    0,                               // no accumulation buffer 
    0, 0, 0, 0,                      // accum bits ignored 
    32,                              // 32-bit z-buffer 
    0,                               // no stencil buffer 
    0,                               // no auxiliary buffer 
    0, //PFD_MAIN_PLANE,                  // main layer 
    0,                               // reserved 
    0, 0, 0                          // layer masks ignored 
  }; 
  int pixelFormat;
  int mouseX  =0 ;
  int mouseY  =0 ;
  int mouseB  =0 ;
  int ACS_stat=0 ;

  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      this.wm_destroy() ;
      wglDeleteContext( hRC );
      ReleaseDC(hwnd,WinDC);
      break ;
      
    case WM_CREATE:
      WinDC=GetDC(hwnd);
      pixelFormat = ChoosePixelFormat(WinDC,&pfd);
      SetPixelFormat(WinDC,pixelFormat,&pfd);
      hRC=wglCreateContext(WinDC);
      wglMakeCurrent( WinDC, hRC );
      this.wm_create() ;
      wglMakeCurrent( null, null );
      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps ;
      HDC         hdc;
      wglMakeCurrent( WinDC, hRC );
      this.wm_paint() ;
      SwapBuffers(WinDC);
      wglMakeCurrent( null, null );
      hdc = BeginPaint(hwnd, &ps);
      EndPaint(hwnd, &ps);
      break ;

    case WM_SIZE:
      wglMakeCurrent( WinDC, hRC );
      this.wm_size(winW,winH) ;
      wglMakeCurrent( null, null );
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      {
	POINT point;

	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );
	mouseX = point.x ;
	mouseY = point.y ;
	SetFocus(getMainWinHWND()) ;
	if( msg==WM_LBUTTONDOWN || msg==WM_RBUTTONDOWN ) SetCapture( hwnd ) ;
	if( msg==WM_LBUTTONUP   || msg==WM_RBUTTONUP   ) ReleaseCapture() ;
	if( msg==WM_MOUSEMOVE ){
	}else{
	  mouseB=0 ;
	  switch( msg ){
	  case WM_LBUTTONDOWN:
	    mouseB|=LBUTTONDOWN ;
	    break ;
	  case WM_LBUTTONUP:
	    mouseB|=LBUTTONUP ;
	    break ;
	  case WM_LBUTTONDBLCLK:
	    mouseB|=LBUTTONDBLCLK ;
	    break ;
	  case WM_RBUTTONDOWN:
	    mouseB|=RBUTTONDOWN ;
	    break ;
	  case WM_RBUTTONUP:
	    mouseB|=RBUTTONUP ;
	    break ;
	  case WM_RBUTTONDBLCLK:
	    mouseB|=RBUTTONDBLCLK ;
	    break ;
          default : break ;
	  }	  
	  
	  if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	  else                            ACS_stat &= ~(SHIFT_KEY) ;
	  if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	  else                            ACS_stat &= ~(CNTL_KEY)  ;
	  if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	  else                            ACS_stat &= ~(ALT_KEY)   ;
	  
	}
	if( this.mouse(ACS_stat, mouseB, mouseX, mouseY )!=0 ){
	  RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	}
      }
      break ;
      
    case WM_CHAR:
      {
	POINT point;
	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );
	if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	else                            ACS_stat &= ~(SHIFT_KEY) ;
	if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	else                            ACS_stat &= ~(CNTL_KEY)  ;
	if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	else                            ACS_stat &= ~(ALT_KEY)   ;
	if( keyboard( 0, ACS_stat, LOBYTE(cast(ushort)(wp)), point.x, point.y )!=0 ){
	  RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	}
      }
      break ;
      
    case WM_KEYDOWN:
      {
	POINT point;
	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );
	if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	else                            ACS_stat &= ~(SHIFT_KEY) ;
	if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	else                            ACS_stat &= ~(CNTL_KEY)  ;
	if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	else                            ACS_stat &= ~(ALT_KEY)   ;
	if( keyboard( 1,ACS_stat, LOBYTE(cast(ushort)(wp)), point.x, point.y )!=0 ){
	  RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	}
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int redraw(){
    if( getHWND()!=null ){
      return( RedrawWindow( getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW) ) ;
    }
    return(-1) ;
  }

  int wm_destroy(){
    return(0) ;
  }
  int wm_create(){
    return(0) ;
  }
  int wm_paint(){
    return(0) ;
  }
  int wm_size(int w, int h){
    return(0) ;
  }

  int mouse( int acs_stat, int mouseb, int mousex, int mousey ){
    return(0) ;
  }

  int keyboard( int tyep, int acs, int key, int mousex, int mousey ){
    return(0) ;
  }

  void setCurrent(){
    wglMakeCurrent( WinDC, hRC );
  }
  
  void unsetCurrent(){
    wglMakeCurrent( null, null );
  }

}
