import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.glut;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import Vec3D;
import Camera;
import OpenGLBase;
import HairDB;
import WorkSpace;
import OpManager;

enum CUROP{
  NONE,
  CAMERA,
  MOVE,
  MOVESEL,
  SCALE,
  SCALESEL
} ;

class OpenGLFrame : OpenGLBase {
  Camera    camera ;
  WorkSpace workspace ;
  int       mouse_Button ;
  int       prev_mouse_Button ;
  int       prev_x ;
  int       prev_y ;
  int       start_x ;
  int       start_y ;
  HairDB    hairdb ;
  OpManager opman  ;
  HMOP      cur_mode = HMOP.SEARCH ;
  bool      scalemode=false ;
  bool      cntlkey_locked=false ;
  CUROP     curop = CUROP.NONE ;
  
  this(string name, string ClassName) {
    super(name,ClassName) ;
    camera    = new Camera() ;
    workspace = new WorkSpace() ;
  }
  
  override int wm_create(){
    workspace.init() ;
    return(0) ;
  }
  
  override int wm_paint(){
    workspace.draw(camera,hairdb) ;
    return(0) ;
  }
  
  override int wm_size(int w, int h){
    camera.win_w=w ; camera.win_h=h ;
    workspace.resize(camera) ;
    if( hairdb !is null ){
      opman.updateDB( HMOP.REDRAW, prev_x, prev_y, 0, 0 ) ;
    }
    return(0) ;
  }
  
  override int mouse( int acs_stat, int mb, int x, int y ){
    int alt_stat   = acs_stat & ALT_KEY   ;
    int shift_stat = acs_stat & SHIFT_KEY ;
    int cntl_stat  = acs_stat & CNTL_KEY  ;

    if( (mb & LBUTTONDOWN)!=0 ) mouse_Button |= 1 ;
    if( (mb & RBUTTONDOWN)!=0 ) mouse_Button |= 2 ;
    if( (mb & LBUTTONUP  )!=0 ) mouse_Button &= ~(1) ;
    if( (mb & RBUTTONUP  )!=0 ) mouse_Button &= ~(2) ;
    
    int rc=0 ;
    if( curop==CUROP.NONE ){
      if( (prev_mouse_Button & 1)==0 && (mouse_Button & 1)!=0 ){  // L-button off => on
	if( alt_stat!=0 ){
	  curop=CUROP.CAMERA ;
	}else{
	  if( scalemode==false ){
	    if( shift_stat==0 && (cntl_stat!=0 || cntlkey_locked==true) ){
	      if( opman.checkSelected(x, y, prev_x, prev_y)<=0 ){
		cur_mode=HMOP.SELECTING ;
		rc=opman.updateDB( HMOP.SELECTINI, x, y, prev_x, prev_y ) ;
	      }else{
		cur_mode=HMOP.UNSELECTING ;
		rc=opman.updateDB( HMOP.UNSELECTINI, x, y, prev_x, prev_y ) ;
	      }
	      curop=CUROP.MOVESEL ;
	    }else if( shift_stat==0 && (cntl_stat==0 && cntlkey_locked==false) ){
	      rc=opman.updateDB( HMOP.MOVEINI, x, y, prev_x, prev_y ) ;
	      curop=CUROP.MOVE ;
	    }
	  }else{
	    if( shift_stat==0 && cntl_stat!=0 ){
	      if( opman.checkSelected(x, y, prev_x, prev_y)<=0 ){
		cur_mode=HMOP.SELECTING ;
		rc=opman.updateDB( HMOP.SELECTINI, x, y, prev_x, prev_y ) ;
	      }else{
		cur_mode=HMOP.UNSELECTING ;
		rc=opman.updateDB( HMOP.UNSELECTINI, x, y, prev_x, prev_y ) ;
	      }
	      curop=CUROP.SCALESEL ;
	    }else if( shift_stat==0 && cntl_stat==0 ){
	      rc=opman.updateDB( HMOP.SCALEINI, x, y, x, y ) ;
	      start_x=x ;
	      start_y=y ;
	      curop=CUROP.SCALE ;
	    }
	  }
	}
      }else{
	rc=opman.updateDB( HMOP.SEARCH, x, y, prev_x, prev_y ) ;
      }
    }else{
      if( (prev_mouse_Button & 1)!=0 && (mouse_Button & 1)==0 ){  // L-button on  => off
	if( scalemode==true ){
	  switch( curop ){
	  case CUROP.SCALE :
	    rc=opman.updateDB( HMOP.SCALEFIX, x, y, start_x, start_y ) ;
	    break ;
	  case CUROP.SCALESEL :
	    if(       cur_mode==HMOP.SELECTING ){
	      rc=opman.updateDB( HMOP.SELECTFIX, x, y, prev_x, prev_y ) ;
	    }else if( cur_mode==HMOP.UNSELECTING ){
	      rc=opman.updateDB( HMOP.UNSELECTFIX, x, y, prev_x, prev_y ) ;
	    }
	    break ;
	  default : break ;
	  }
	}else{
	  switch( curop ){
	  case CUROP.MOVE :
	    rc=opman.updateDB( HMOP.MOVEFIX, x, y, prev_x, prev_y ) ;
	    break ;
	  case CUROP.MOVESEL :
	    if(       cur_mode==HMOP.SELECTING ){
	      rc=opman.updateDB( HMOP.SELECTFIX, x, y, prev_x, prev_y ) ;
	    }else if( cur_mode==HMOP.UNSELECTING ){
	      rc=opman.updateDB( HMOP.UNSELECTFIX, x, y, prev_x, prev_y ) ;
	    }
	    break ;
	  default : break ;
	  }
	}
	curop=CUROP.NONE ;
	rc=opman.updateDB( HMOP.SEARCH, x, y, prev_x, prev_y ) ;
      }else{
	switch( curop ){
	case CUROP.CAMERA :
	  if( alt_stat!=0 ){
	    if(  shift_stat==0 && cntl_stat==0 ){
	      camera.addRotY( x-prev_x ) ;
	      camera.addRotX( y-prev_y ) ;
	    }else if( shift_stat==0 && cntl_stat!=0 ){
	      camera.addEyeZ( cast(double)(prev_y-y) ) ;
	    }else if( shift_stat!=0 && cntl_stat==0 ){
	      camera.addShift(x,y,prev_x,prev_y) ;
	    }
	  }
	  rc=opman.updateDB( HMOP.REDRAW, x, y, prev_x, prev_y ) ;
	  break ;
	case CUROP.MOVE :
	  rc=opman.updateDB( HMOP.MOVING, x, y, prev_x, prev_y ) ;
	  break ;
	case CUROP.MOVESEL :
	  rc=opman.updateDB( cur_mode, x, y, prev_x, prev_y ) ;
	  break ;
	case CUROP.SCALE :
	  rc=opman.updateDB( HMOP.SCALING, x, y, start_x, start_y ) ;
	  break ;
	case CUROP.SCALESEL :
	  rc=opman.updateDB( cur_mode, x, y, prev_x, prev_y ) ;
	  break ;
	default : core.exception.onSwitchError() ;
	}
      }
    }
    prev_x=x ;
    prev_y=y ;
    prev_mouse_Button=mouse_Button ;
    return( rc ) ;
  }

  override int keyboard( int type, int acs, int key, int x, int y){
    return( opman.keyboard(type,acs,key,x,y) ) ;
  }

  int setCntlKeyLocked(bool locked){
    cntlkey_locked=locked ;
    return(0) ;
  }

  int setScaleMode(bool mode){
    scalemode=mode ;
    return(0) ;
  }
  
  int setHairDB(HairDB hairdb){
    if( hairdb is null ) return(-1) ;
    this.hairdb=hairdb ;
    this.hairdb.setCamera(camera) ;
    return(0) ;
  }

  int setOpManager(OpManager opman){
    if( opman is null ) return(-1) ;
    this.opman=opman ;
    setHairDB(this.opman.getHairDB()) ;
    return(0) ;
  }

  WorkSpaceEnv getWorkSpaceEnv(){
    return(workspace.getWorkSpaceEnv()) ;
  }

  int setWorkSpaceEnv(WorkSpaceEnv wse){
    return(workspace.setWorkSpaceEnv(wse)) ;
  }
  
}
