import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.DialogBoxClass;
import win32frame.PushButton;
import win32frame.TextBox;
import util.BitMap;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import Spin;
import MyButton ;
import ColorSelectDialog;

import WorkSpace ;

COLORREF RGB(uint r, uint g, uint b) {
  return cast(COLORREF)(r | (g << 8) | (b << 16));
}
COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

enum CMD{
  BGCOL,
  PTRCOL,
  OBJDIFF,
  OBJAMB,
  OBJEDGE,
  OK,
  CANCEL
} ;

struct ColorBtnList{
  MyButton wobj ;
  string   label ;
  HMENU    cmd ;
  uint     color ;
  int      img_w=80 ;
  int      img_h=20 ;

  void set(string label, HMENU cmd, double r, double g, double b){
    this.label=label ;
    this.cmd  =cmd ;
    this.color=RGBA(cast(uint)(round(255.0*r)),cast(uint)(round(255.0*g)),cast(uint)(round(255.0*b)),255) ;
  }
} ;

class PrefDialog : DialogBoxClass {
  ColorBtnList[] colbtnlist ;
  
  PushButton   buttonOK  ;
  PushButton   buttonCAN ;
  HFONT        hfont  ;

  WorkSpaceEnv wse ;
  int          dlgstate  ;

  const int    defaultWidth =200 ;
  const int    defaultHeight=300 ;
  const DWORD  defaultStyle = WS_VISIBLE | WS_SYSMENU ;

  this(string name, string ClassName) {
    UnregisterClass( toUTFz!(wchar*)(ClassName), GetModuleHandle(null) ) ;
    winc.cbSize         = WNDCLASSEX.sizeof;
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = null ;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
  }

  int Create( int x, int y, HWND phwnd, HMENU hmenu){
    return( create(defaultStyle,x,y,defaultWidth,defaultHeight,phwnd,hmenu) ) ;
  }
  
  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont);
      break ;
    case WM_CREATE:
      hfont = CreateFont(
			 12 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
			 DEFAULT_CHARSET, OUT_DEFAULT_PRECIS ,
			 CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY , 
			 FIXED_PITCH, null
			 );

      int btnnum=5 ;
      colbtnlist = new ColorBtnList[btnnum] ;

      colbtnlist[0].set("Background",      cast(HMENU)(CMD.BGCOL),   wse.bg_r      , wse.bg_g      , wse.bg_b     ) ;
      colbtnlist[1].set("Pointer",         cast(HMENU)(CMD.PTRCOL),  wse.point_r   , wse.point_g   , wse.point_b  ) ;
      colbtnlist[2].set("Object Diffuse",  cast(HMENU)(CMD.OBJDIFF), wse.objdiff_r , wse.objdiff_g , wse.objdiff_b) ;
      colbtnlist[3].set("Object Ambient",  cast(HMENU)(CMD.OBJAMB),  wse.objamb_r  , wse.objamb_g  , wse.objamb_b ) ;
      colbtnlist[4].set("Object EdgeLine", cast(HMENU)(CMD.OBJEDGE), wse.objedge_r , wse.objedge_g , wse.objedge_b) ;
      
      foreach( i, ref colbtn ; colbtnlist ){
	auto colimg = new BitMap() ;
	colimg.setImage(makeColorButtonFace(colbtn.color,colbtn.img_w,colbtn.img_h),colbtn.img_w,colbtn.img_h,false,false) ;
	colbtn.wobj = new MyButton(colbtn.label) ;
	colbtn.wobj.setBitMap(colimg) ;
	colbtn.wobj.addStyle(WS_TABSTOP) ;
	colbtn.wobj.setPosition(100, 8+cast(int)(i)*20,80,20) ;
	addObject(colbtn.wobj, colbtn.cmd, false, false) ;
      }

      buttonOK= new PushButton("OK") ;
      buttonOK.addStyle(WS_TABSTOP) ;
      buttonOK.setPosition(winW-184,winH-32,80,20) ;
      addObject( buttonOK, cast(HMENU)(CMD.OK) ) ;
      buttonOK.setFont(hfont) ;

      buttonCAN= new PushButton("Cancel") ;
      buttonCAN.addStyle(WS_TABSTOP) ;
      buttonCAN.setPosition(winW-92,winH-32,80,20) ;
      addObject( buttonCAN, cast(HMENU)(CMD.CANCEL) ) ;
      buttonCAN.setFont(hfont) ;

      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC hdc = BeginPaint(hwnd, &ps);
      SetBkMode(hdc , TRANSPARENT);
      foreach( i, colbtn ; colbtnlist ){
	SIZE labelsize ;
	wchar[] label=(toUTF16(colbtn.label)).dup ;

	SelectObject(hdc,hfont) ;
	GetTextExtentPoint32( hdc, cast(wchar*)(label), cast(int)(label.length), &labelsize);
	TextOut(hdc,
		8,8+4+cast(int)(i)*20,
		cast(wchar*)(label), cast(int)(label.length)) ;
      }
      EndPaint(hwnd, &ps);
      break ;
      
    case WM_SIZE:
      break ;
      
    case WM_CLOSE:
      setResult(0) ;
      DestroyWindow(hwnd);
      break ;
      
    case WM_COMMAND:
      dlgstate=LOWORD(wp) ;
      switch( dlgstate ){
      case CMD.BGCOL:
      case CMD.PTRCOL:
      case CMD.OBJDIFF:
      case CMD.OBJAMB:
      case CMD.OBJEDGE:
	RECT wrect ;
        GetWindowRect(this.getMainWinHWND(),&wrect) ;
	foreach( i, ref colbtn ; colbtnlist ){
	  if( cast(int)(colbtn.cmd)==dlgstate ){
	    auto colsel = new ColorSelectDialog("","colsel") ;
	    colsel.setColor(colbtn.color) ;
	    colsel.enableAlpha(false) ;
	    colsel.Create( wrect.left+120,wrect.top+120,hwnd,null ) ;
	    if( colsel.getDlogResult()==1 ){
	      colbtn.color=colsel.getColor() ;
	      auto colimg = new BitMap() ;
	      colimg.setImage(makeColorButtonFace(colbtn.color,colbtn.img_w,colbtn.img_h),colbtn.img_w,colbtn.img_h,false,false) ;
	      colbtn.wobj.setBitMap(colimg) ;
	      RedrawWindow( colbtn.wobj.getHWND(), null, null, RDW_INVALIDATE ) ;
	    }
	    break ;
	  }
	}
        break ;

      case CMD.CANCEL:
        setResult(0) ;
        DestroyWindow(hwnd);
        break ;
      case CMD.OK:
        setResult(1) ;
        DestroyWindow(hwnd);
        break ;
	
      default :
	break ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int setDefaultWorkSpaceEnv(WorkSpaceEnv wse){
    this.wse=wse ;
    return(0) ;
  }
  
  WorkSpaceEnv getWorkSpaceEnv(){
    wse.bg_r      = cast(double)((colbtnlist[0].color    )&0x000000ff)/255.0 ;
    wse.bg_g      = cast(double)((colbtnlist[0].color>>8 )&0x000000ff)/255.0 ;
    wse.bg_b      = cast(double)((colbtnlist[0].color>>16)&0x000000ff)/255.0 ;
    wse.point_r   = cast(double)((colbtnlist[1].color    )&0x000000ff)/255.0 ;
    wse.point_g   = cast(double)((colbtnlist[1].color>>8 )&0x000000ff)/255.0 ;
    wse.point_b   = cast(double)((colbtnlist[1].color>>16)&0x000000ff)/255.0 ;
    wse.objdiff_r = cast(double)((colbtnlist[2].color    )&0x000000ff)/255.0 ;
    wse.objdiff_g = cast(double)((colbtnlist[2].color>>8 )&0x000000ff)/255.0 ;
    wse.objdiff_b = cast(double)((colbtnlist[2].color>>16)&0x000000ff)/255.0 ;
    wse.objamb_r  = cast(double)((colbtnlist[3].color    )&0x000000ff)/255.0 ;
    wse.objamb_g  = cast(double)((colbtnlist[3].color>>8 )&0x000000ff)/255.0 ;
    wse.objamb_b  = cast(double)((colbtnlist[3].color>>16)&0x000000ff)/255.0 ;
    wse.objedge_r = cast(double)((colbtnlist[4].color    )&0x000000ff)/255.0 ;
    wse.objedge_g = cast(double)((colbtnlist[4].color>>8 )&0x000000ff)/255.0 ;
    wse.objedge_b = cast(double)((colbtnlist[4].color>>16)&0x000000ff)/255.0 ;
    return(wse) ;
  }

  uint[] makeColorButtonFace(uint col, int w, int h){
    uint[] img = new uint[w*h] ;
    for( int y=0 ; y<h ; y++ ){
      for( int x=0 ; x<w ; x++ ){
	int bgv = ((x/8+y/8)%2)==0 ? 128 : 70 ;
	int a = (col>>24) & 0x000000ff ;
	int r = (col    ) & 0x000000ff ;
	int g = (col>>8 ) & 0x000000ff ;
	int b = (col>>16) & 0x000000ff ;
	r = (r*a + bgv*(255-a)) /255 ;
	g = (g*a + bgv*(255-a)) /255 ;
	b = (b*a + bgv*(255-a)) /255 ;
        img[y*w+x]=RGBA(b,g,r,255) ;
      }
    }
    return(img) ;
  }
    
}
  
