import std.stdio;
import std.math;

struct vec2D  {
  double  u ;
  double  v ;
} ;

struct vec3D  {
  double  x  ;
  double  y  ;
  double  z  ;
  int     id ;
  bool    sel  ;
  bool    disp ;

  void set(double x, double y, double z){
    this.x = x ;
    this.y = y ;
    this.z = z ;
  }

  void opAddAssign(vec3D d){
    this.x += d.x ;
    this.y += d.y ;
    this.z += d.z ;
  }
  
  void opAddAssign(double d){
    this.x += d ;
    this.y += d ;
    this.z += d ;
  }
  
  vec3D opAdd(vec3D d){
    vec3D r ;
    r.x = this.x + d.x ;
    r.y = this.y + d.y ;
    r.z = this.z + d.z ;
    return(r) ;
  }
  
  vec3D opAdd(double d){
    vec3D r ;
    r.x = this.x + d ;
    r.y = this.y + d ;
    r.z = this.z + d ;
    return(r) ;
  }
  
  void opSubAssign(vec3D d){
    this.x -= d.x ;
    this.y -= d.y ;
    this.z -= d.z ;
  }

  void opSubAssign(double d){
    this.x -= d ;
    this.y -= d ;
    this.z -= d ;
  }
  
  vec3D opSub(vec3D d){
    vec3D r ;
    r.x = this.x - d.x ;
    r.y = this.y - d.y ;
    r.z = this.z - d.z ;
    return(r) ;
  }
  
  vec3D opSub(double d){
    vec3D r ;
    r.x = this.x - d ;
    r.y = this.y - d ;
    r.z = this.z - d ;
    return(r) ;
  }
  
  void opMulAssign(double d){
    this.x *= d ;
    this.y *= d ;
    this.z *= d ;
  }
  
  vec3D opMul(double d){
    vec3D r ;
    r.x = this.x * d ;
    r.y = this.y * d ;
    r.z = this.z * d ;
    return(r) ;
  }
  
  void opDivAssign(double d){
    this.x /= d ;
    this.y /= d ;
    this.z /= d ;
  }
  
  vec3D opDiv(double d){
    vec3D r ;
    r.x = this.x / d ;
    r.y = this.y / d ;
    r.z = this.z / d ;
    return(r) ;
  }
  
  vec3D opNeg(){
    vec3D r ;
    r.x = -this.x ;
    r.y = -this.y ;
    r.z = -this.z ;
    return(r) ;
  }
  
  const bool opEquals( ref const(vec3D) b ){
    if( this.x == b.x &&
	this.y == b.y &&
	this.z == b.z ) {
      return(true) ;
    }
    return(false) ;
  }
  
  double distance( vec3D b ){
    double lx=this.x-b.x ;
    double ly=this.y-b.y ;
    double lz=this.z-b.z ;
    return( sqrt( lx*lx + ly*ly + lz*lz ) ) ;
  }
  
  double length(){
    return( sqrt( this.x*this.x + this.y*this.y + this.z*this.z ) ) ;
  }
  
  vec3D normalize(){
    vec3D  r ;
    double d ;
    d=sqrt( this.x*this.x + this.y*this.y + this.z*this.z ) ;
    if( d==0.0 ){
      r.x=0.0 ;
      r.y=0.0 ;
      r.z=0.0 ;
    }else{
      r.x = this.x/d ;
      r.y = this.y/d ;
      r.z = this.z/d ;
    }
    return(r) ;
  }

  vec3D rotate( double angle, vec3D vec ){
    double sinval, cosval;
    double m[3][3] ;
    double x,y,z ;
    vec3D  r ;
    
    x=vec.x ;
    y=vec.y ;
    z=vec.z ;
    sinval = sin(angle);
    cosval = cos(angle);
    
    m[0][0] = x * x * (1.0 - cosval) + cosval  ;
    m[0][1] = x * y * (1.0 - cosval) - z * sinval;
    m[0][2] = x * z * (1.0 - cosval) + y * sinval;
    
    m[1][0] = x * y * (1.0 - cosval) + z * sinval;
    m[1][1] = y * y * (1.0 - cosval) + cosval  ;
    m[1][2] = y * z * (1.0 - cosval) - x * sinval;
    
    m[2][0] = x * z * (1.0 - cosval) - y * sinval;
    m[2][1] = y * z * (1.0 - cosval) + x * sinval;
    m[2][2] = z * z * (1.0 - cosval) + cosval  ;
    
    r.x=m[0][0]*this.x + m[0][1]*this.y + m[0][2]*this.z ;
    r.y=m[1][0]*this.x + m[1][1]*this.y + m[1][2]*this.z ;
    r.z=m[2][0]*this.x + m[2][1]*this.y + m[2][2]*this.z ;
    
    return(r);
  }
  
  double dot( vec3D b ){
    return( (this.x * b.x) + (this.y * b.y) + (this.z * b.z) ) ;
  }
  
  vec3D cross( vec3D b ){
    double tx,ty,tz ;
    vec3D  r ;
    r.x=(this.y * b.z) - (this.z * b.y) ;
    r.y=(this.z * b.x) - (this.x * b.z) ;
    r.z=(this.x * b.y) - (this.y * b.x) ;
    return(r) ;
  }
  
  bool LPcross( vec3D eyepos, vec3D ray, vec3D fp0, vec3D fn) {
    double  a, b, t;

    b = ray.dot(fn) ;
    if( b == 0.0 ) return(false);
    
    fp0-=eyepos ;
    a = fp0.dot(fn) ;
    t = a/b ;
    
    this.x = eyepos.x + ray.x * t ;
    this.y = eyepos.y + ray.y * t ;
    this.z = eyepos.z + ray.z * t ;
    
    return(true);
  }
  
  bool chkinside( vec3D fp0, vec3D fp1, vec3D fp2 ) {
    vec3D   n0 ;
    vec3D   n1 ;
    vec3D   n2 ;
    double  res0,res1,res2;
    
    fp0.x-=this.x ; fp0.y-=this.y ; fp0.z-=this.z ;
    fp1.x-=this.x ; fp1.y-=this.y ; fp1.z-=this.z ;
    fp2.x-=this.x ; fp2.y-=this.y ; fp2.z-=this.z ;
    n0=fp0 ;
    n1=fp1 ;
    n2=fp2 ;
    n0=n0.cross(fp1) ;
    n1=n1.cross(fp2) ;
    n2=n2.cross(fp0) ;
    res0=n0.dot(n1) ;
    res1=n1.dot(n2) ;
    res2=n2.dot(n0) ;
    res0 = res0<0.0 ? -1.0 : 1.0 ;
    res1 = res1<0.0 ? -1.0 : 1.0 ;
    res2 = res2<0.0 ? -1.0 : 1.0 ;
    
    if( res0==res1 && res1==res2 && res2==res0 ){
      return(true) ;
    }
    return(false) ;
  }

  void print() {
    writef("x=%f, y=%f, z=%f\n",x,y,z) ;
  }
} ;

struct face3D {
  int    p0 ;
  int    p1 ;
  int    p2 ;
  int    id ;

  void set(int p0, int p1, int p2){
    this.p0=p0 ;
    this.p1=p1 ;
    this.p2=p2 ;
  }
} ;

