import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.glut;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import Vec3D;
import Camera;
import HairDB;

struct WorkSpaceEnv{
  double bg_r      = 0.0 ;
  double bg_g      = 0.0 ;
  double bg_b      = 0.7 ;
  double point_r   = 1.0 ;
  double point_g   = 0.0 ;
  double point_b   = 0.0 ;
  double objdiff_r = 0.8 ;
  double objdiff_g = 0.8 ;
  double objdiff_b = 0.8 ;
  double objamb_r  = 0.5 ;
  double objamb_g  = 0.5 ;
  double objamb_b  = 0.5 ;
  double objedge_r = 0.0 ;
  double objedge_g = 0.0 ;
  double objedge_b = 0.0 ;
} ;

class WorkSpace {
  WorkSpaceEnv wse ;
  
  this() {
  }
  
  int init(){
    glClearColor(wse.bg_r,wse.bg_g,wse.bg_b,0.0) ;
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    return(0) ;
  }
  
  int draw( Camera camera, HairDB hairdb ){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    
    glViewport(0, 0, camera.win_w, camera.win_h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(camera.fovy, cast(double)(camera.win_w)/cast(double)(camera.win_h), 0.1, 10000.0);
    gluLookAt(camera.eye.x, camera.eye.y, camera.eye.z,
	      camera.cnt.x, camera.cnt.y, camera.cnt.z,
	      0.0,1.0, 0.0) ;
    glMatrixMode(GL_MODELVIEW);
    
    glClearColor(wse.bg_r,wse.bg_g,wse.bg_b,0.0) ;
    
    glTranslated( camera.pan.x , camera.pan.y , camera.pan.z);
    glRotated( camera.rot.x , 1.0, 0.0, 0.0 );
    glRotated( camera.rot.y , 0.0, 1.0, 0.0 );
    glRotated( camera.rot.z , 0.0, 0.0, 1.0 );
    
    // Draw mouse pointer
    glDisable(GL_LIGHTING);
    glPointSize(0.5);
    glEnable(GL_POINT_SMOOTH);
    glColor3d(wse.point_r, wse.point_g, wse.point_b);
    glBegin(GL_POINTS);
    glVertex3d(camera.cursor3D.x, camera.cursor3D.y, camera.cursor3D.z) ;
    glEnd();
    
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 1.0) ;
    
    if( hairdb !is null ){
      EDMD cur_mode = hairdb.getEditMode() ;
      SearchST searchID = hairdb.getSearchID() ;
      
      // Draw HeadObject
      vec3D[]  HRvtx  = hairdb.getHRvtx()  ;
      face3D[] HRface = hairdb.getHRface() ;
      vec3D[]  HRfnml = hairdb.getHRfnml() ;
      
      if( hairdb.getHideHRface()==false ){
	GLfloat[4] mat_diffuse=[wse.objdiff_r,wse.objdiff_g,wse.objdiff_b,1.0] ;
	GLfloat[4] mat_ambient=[wse.objamb_r, wse.objamb_g, wse.objamb_b, 1.0] ;
	glMaterialfv(GL_FRONT, GL_DIFFUSE, &mat_diffuse[0]);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, &mat_ambient[0] ) ;
	foreach( i, f ; HRface ){
	  glBegin( GL_TRIANGLES ) ;
	  glNormal3d(HRfnml[i].x,HRfnml[i].y,HRfnml[i].z);
	  glVertex3d(HRvtx[f.p0].x,HRvtx[f.p0].y,HRvtx[f.p0].z) ;
	  glVertex3d(HRvtx[f.p1].x,HRvtx[f.p1].y,HRvtx[f.p1].z) ;
	  glVertex3d(HRvtx[f.p2].x,HRvtx[f.p2].y,HRvtx[f.p2].z) ;
	  glEnd();
	}
	glDisable(GL_LIGHTING);
	glColor3d(wse.objedge_r, wse.objedge_g, wse.objedge_b);
	foreach( i, f ; HRface ){
	  glBegin( GL_LINE_LOOP ) ;
	  glNormal3d(HRfnml[i].x,HRfnml[i].y,HRfnml[i].z);
	  glVertex3d(HRvtx[f.p0].x,HRvtx[f.p0].y,HRvtx[f.p0].z) ;
	  glVertex3d(HRvtx[f.p1].x,HRvtx[f.p1].y,HRvtx[f.p1].z) ;
	  glVertex3d(HRvtx[f.p2].x,HRvtx[f.p2].y,HRvtx[f.p2].z) ;
	  glEnd();
	}
      }

      // Draw HairGuide Surface
      int[] suflist=hairdb.getSurfaceList() ;
      if( suflist.length>=2 ){
	GLfloat[4] mat_diffuse=[0.6,0.6,0.8,1.0] ;
	GLfloat[4] mat_ambient=[0.5,0.5,0.5,1.0] ;
	glMaterialfv(GL_FRONT, GL_DIFFUSE, &mat_diffuse[0]);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, &mat_ambient[0] ) ;
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);
	GLUnurbsObj *nurb = gluNewNurbsRenderer();
	gluNurbsProperty(nurb, GLU_SAMPLING_TOLERANCE, 50.0);
	gluNurbsProperty(nurb, GLU_DISPLAY_MODE, GLU_FILL);
	HairGuide[] HGlist ;
	HGlist = hairdb.getFloatHGlist() ;
	if( HGlist.length==0 ){
	  HGlist=hairdb.getHGlist() ;
	}
	for( int i=1 ; i<suflist.length ; i++ ){
	  float[48] cp ;
	  float[8]  knots = [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0];
	  for( int j=0 ; j<2 ; j++ ){
	    for( int k=0 ; k<4 ; k++ ){
	      cp[ 0 + k*3+0]=HGlist[suflist[i  ]].plist[k+j*3].x ;
	      cp[ 0 + k*3+1]=HGlist[suflist[i  ]].plist[k+j*3].y ;
	      cp[ 0 + k*3+2]=HGlist[suflist[i  ]].plist[k+j*3].z ;
	      cp[12 + k*3+0]=HGlist[suflist[i  ]].plist[k+j*3].x ;
	      cp[12 + k*3+1]=HGlist[suflist[i  ]].plist[k+j*3].y ;
	      cp[12 + k*3+2]=HGlist[suflist[i  ]].plist[k+j*3].z ;
	      cp[24 + k*3+0]=HGlist[suflist[i-1]].plist[k+j*3].x ;
	      cp[24 + k*3+1]=HGlist[suflist[i-1]].plist[k+j*3].y ;
	      cp[24 + k*3+2]=HGlist[suflist[i-1]].plist[k+j*3].z ;
	      cp[36 + k*3+0]=HGlist[suflist[i-1]].plist[k+j*3].x ;
	      cp[36 + k*3+1]=HGlist[suflist[i-1]].plist[k+j*3].y ;
	      cp[36 + k*3+2]=HGlist[suflist[i-1]].plist[k+j*3].z ;
	    }
	    gluBeginSurface(nurb);
	    gluNurbsSurface(nurb, 
			    8, knots.ptr,
			    8, knots.ptr,
			    4 * 3,
			    3,
			    cp.ptr, 
			    4, 4,
			    GL_MAP2_VERTEX_3);
	    gluEndSurface(nurb);
	  }
	}
	gluDeleteNurbsRenderer(nurb) ;
      }
      
      glDisable( GL_LIGHTING );
      
      // Draw HairRoot
      if( hairdb.getHideHRvtx()==false ){
	glPointSize(5.0);
	glBegin(GL_POINTS);
	foreach( i, v ; HRvtx ){
	  if( v.disp==false ) continue ;
	  if( cur_mode==EDMD.HR && i==searchID.id0 ){
	    glColor3d(1.0, 0.0, 0.0);
	  }else{
	    if( v.sel==true ){
	      glColor3d(0.5, 0.0, 0.0);
	    }else{
	      glColor3d(1.0, 1.0, 1.0);
	    }
	  }
	  glVertex3d(v.x, v.y, v.z) ;
	}
	glEnd();
      }
      
      // Draw HairGuide line
      HairGuide[] HGlist ;
      HGlist = hairdb.getFloatHGlist() ;
      if( HGlist.length==0 ){
	HGlist=hairdb.getHGlist() ;
      }
      GIDSTATE[] gid_statlist = hairdb.getGIDstatlist() ;
      GIDDISP[]  gid_displist = hairdb.getGIDdisplist() ;
      float[8]           knots ;
      float[POINT_NUM*3] plist ;

      foreach( ref gid_disp ; gid_displist ){
	gid_disp.r = cast(double)( gid_disp.color      & 0x000000ff)/255.0 ;
	gid_disp.g = cast(double)((gid_disp.color>>8 ) & 0x000000ff)/255.0 ;
	gid_disp.b = cast(double)((gid_disp.color>>16) & 0x000000ff)/255.0 ;
      }
      
      if( hairdb.getHideHGline()==false ){
	GLUnurbsObj *nurb = gluNewNurbsRenderer();
	gluNurbsProperty(nurb, GLU_SAMPLING_TOLERANCE, 25.0);
	gluNurbsProperty(nurb, GLU_DISPLAY_MODE, GLU_FILL);
	foreach( i, hg ; HGlist ){
	  if( gid_statlist[hg.gid].hide==true ) continue ;
	  for( int j=0 ; j<knots.length ; j++ ){
	    knots[j]=cast(float)(hg.knots[j]) ;
	  }
	  for( int j=0,k=0 ; j<plist.length ; j+=3,k++ ){
	    plist[j+0]=cast(float)(hg.plist[k].x) ;
	    plist[j+1]=cast(float)(hg.plist[k].y) ;
	    plist[j+2]=cast(float)(hg.plist[k].z) ;
	  }
	  if( cur_mode==EDMD.HG && i==searchID.id0 ){
	    glColor3d(1.0, 0.0, 0.0);
	  }else{
	    if( hg.sel==true){
	      glColor3d(0.5, 0.0, 0.0);
	    }else{
	      glColor3d(gid_displist[hg.gid].r,
			gid_displist[hg.gid].g,
			gid_displist[hg.gid].b) ;
	    }
	  }
	  for( int j=0 ; j<((plist.length/3)-1)/3 ; j++ ){
	    gluBeginCurve(nurb);
	    gluNurbsCurve( nurb,
			   knots.length, &knots[0],
			   3,
			   &plist[(j*3)*3],
			   4,
			   GL_MAP1_VERTEX_3) ;
	    gluEndCurve(nurb);
	  }
	}
	gluDeleteNurbsRenderer(nurb) ;
      }
      // Draw HairGuide Control Point
      foreach( i, hg ; HGlist ){
	if( gid_statlist[hg.gid].hide==true ) continue ;
	if( cur_mode==EDMD.HG && i==searchID.id0 ){
	  glColor3d(1.0, 0.0, 0.0);
	}else{
	  if( hg.sel==true){
	    glColor3d(0.5, 0.0, 0.0);
	  }else{
	    glColor3d(gid_displist[hg.gid].r,
		      gid_displist[hg.gid].g,
		      gid_displist[hg.gid].b) ;
	  }
	}
	for( int j=0 ; j<hg.plist.length ; j++ ){
	  if( j%3==0 ) glPointSize(5.0);
	  else         glPointSize(3.0);
	  if( hairdb.getHideHGvtx()==false ||
	     (hairdb.getHideHGvtx()==true && cur_mode==EDMD.HG && i==searchID.id0) ){
	    glBegin(GL_POINTS);
	    glVertex3d(hg.plist[j].x,hg.plist[j].y,hg.plist[j].z) ;
	    glEnd();
	  }
	  if( cur_mode==EDMD.HG && i==searchID.id0 ){
	    switch(j%3){
	    case 1:
	      glBegin(GL_LINES);
	      glVertex3d(hg.plist[j  ].x,hg.plist[j  ].y,hg.plist[j  ].z) ;
	      glVertex3d(hg.plist[j-1].x,hg.plist[j-1].y,hg.plist[j-1].z) ;
	      glEnd();
	      break ;
	    case 2:
	      glBegin(GL_LINES);
	      glVertex3d(hg.plist[j  ].x,hg.plist[j  ].y,hg.plist[j  ].z) ;
	      glVertex3d(hg.plist[j+1].x,hg.plist[j+1].y,hg.plist[j+1].z) ;
	      glEnd();
	      break ;
	    default : break ;
	    }
	  }
	}
      }
    }
    
    glDisable(GL_POLYGON_OFFSET_FILL);
    
    glDisable( GL_LIGHTING );
    // 文字表示
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluOrtho2D( 0.0, cast(double)(camera.win_w), cast(double)(camera.win_h), 0.0  );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glColor3d( 1.0, 1.0, 1.0 ) ;
    
    //glut_print(1,12,format("cursor=%f,%f,%f",camera.cursor3D.x,camera.cursor3D.y,camera.cursor3D.z)) ;
    //glut_print(1,25,format("eye=%f,%f,%f ray=%f,%f,%f",camera.geyepos.x,camera.geyepos.y,camera.geyepos.z,
    //			                               camera.geyeray.x,camera.geyeray.y,camera.geyeray.z)) ;
    
    glFinish();
    return(0) ;
  }
  
  void glut_print(int x, int y, ...){
    void delegate(dchar) putc = &glut_putc ;
    glRasterPos2i(x,y) ;
    std.format.doFormat( putc, _arguments, _argptr ) ;
  }
  
  void glut_putc(dchar c){
    glutBitmapCharacter(cast(void*)(GLUT_BITMAP_8_BY_13), c);
  }
  
  int resize(Camera camera){
    glViewport(0, 0, camera.win_w, camera.win_h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(camera.fovy, cast(double)(camera.win_w)/cast(double)(camera.win_h), 0.1, 10000.0);
    gluLookAt(camera.eye.x, camera.eye.y, camera.eye.z,
	      camera.cnt.x, camera.cnt.y, camera.cnt.z,
	      0.0,1.0, 0.0) ;
    glMatrixMode(GL_MODELVIEW);
    return(0) ;
  }

  WorkSpaceEnv getWorkSpaceEnv(){
    return(wse) ;
  }
  
  int setWorkSpaceEnv(WorkSpaceEnv wse){
    this.wse = wse ;
    return(0) ;
  }
  
}
