import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.Frame;
import win32frame.WindowMenu;
import win32frame.StatusBar;
import win32frame.PushButton;
import win32frame.MainWindow ;
import win32frame.MsgBox;
import win32frame.FileDialog;
import std.stdio;
import std.string;
import std.utf;

import MainPanel ;
import HairDB ;
import OpManager;
import UndoBuffer;
import UndoPack;
import ExportPOVDialog;
import PrefDialog;
import VerInfoDialog;
import OpHelpDialog;

enum MWFILE{
  ROOT   =0x100,
  OPEN   ,
  SAVE   ,
  SAVEAS ,
  IMPORT ,
  EXPORT ,
  EXIT      
} ;
enum MWEDIT{
  ROOT    = 0x200,
  SELALL  ,
  UNSELALL,
} ;

enum MWVIEW{
  ROOT  =0x300,
  RESET ,
  HGCP  ,
  HR    ,
  OBJ
} ;

enum MWVIEWSUB{
  SHOW =0x010,
  HIDE =0x020
} ;

enum MWOPTION{
  ROOT       = 0x400,
  PREFERENCE
} ;

enum MWHELP{
  ROOT     = 0x500,
  OPHELP,  
  VERINFO  
} ;

class myMainWindow : MainWindow {
  WindowMenu   menuroot ;
  StatusBar    statusbar ;
  MainPanel    mainpanel ;

  HairDB       hairdb ;
  OpManager    opman  ;
  UndoBuffer   undobuf ;

  const string appname="HairMaker" ;
  const string hairfilter="HairMaker {*.hair}\0*.hair\0" ;
  const string objfilter ="WavefrontOBJ {*.obj}\0*.obj\0" ;
  const string povfilter ="POV-Ray inc file {*.inc}\0*.inc\0" ;
  
  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = LoadIcon(GetModuleHandle(null), toUTFz!(wchar*)("HMICON"));
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name,ClassName) ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      PostQuitMessage(0);
      break ;

    case WM_CREATE:
      menuroot = new WindowMenu() ;
      menuroot.create() ;
      menuroot.addMenu(             MFT_STRING, "File"         , MWFILE.ROOT   ) ;
      menuroot.addItem( "File",     MFT_STRING, "Open"         , MWFILE.OPEN   ) ;
      menuroot.addItem( "File",     MFT_STRING, "Save"         , MWFILE.SAVE   ) ;
      menuroot.addItem( "File",     MFT_STRING, "SaveAs"       , MWFILE.SAVEAS ) ;
      menuroot.addItem( "File",     MFT_STRING, "Import"       , MWFILE.IMPORT ) ;
      menuroot.addItem( "File",     MFT_STRING, "Export"       , MWFILE.EXPORT ) ;
      menuroot.addItem( "File",     MFT_STRING, "Exit"         , MWFILE.EXIT   ) ;
      menuroot.addMenu(             MFT_STRING, "Edit"         , MWEDIT.ROOT   ) ;
      menuroot.addItem( "Edit",     MFT_STRING, "Select All"   , MWEDIT.SELALL ) ;
      menuroot.addItem( "Edit",     MFT_STRING, "Unselect All" , MWEDIT.UNSELALL) ;
      menuroot.addMenu(             MFT_STRING, "View"         , MWVIEW.ROOT   ) ;
      menuroot.addItem( "View",     MFT_STRING, "Reset"        , MWVIEW.RESET  ) ;
      menuroot.addMenu( "View",     MFT_STRING, "HGcntl"       , MWVIEW.HGCP   ) ;
      menuroot.addItem( "HGcntl",   MFT_STRING, "Show"         , MWVIEW.HGCP|MWVIEWSUB.SHOW ) ;
      menuroot.addItem( "HGcntl",   MFT_STRING, "Hide"         , MWVIEW.HGCP|MWVIEWSUB.HIDE ) ;
      menuroot.addMenu( "View",     MFT_STRING, "HairRoot"     , MWVIEW.HR     ) ;
      menuroot.addItem( "HairRoot", MFT_STRING, "Show"         , MWVIEW.HR|MWVIEWSUB.SHOW ) ;
      menuroot.addItem( "HairRoot", MFT_STRING, "Hide"         , MWVIEW.HR|MWVIEWSUB.HIDE ) ;
      menuroot.addMenu( "View",     MFT_STRING, "Object"       , MWVIEW.OBJ    ) ;
      menuroot.addItem( "Object",   MFT_STRING, "Show"         , MWVIEW.OBJ|MWVIEWSUB.SHOW ) ;
      menuroot.addItem( "Object",   MFT_STRING, "Hide"         , MWVIEW.OBJ|MWVIEWSUB.HIDE ) ;
      menuroot.addMenu(             MFT_STRING, "Option"       , MWOPTION.ROOT ) ;
      menuroot.addItem( "Option",   MFT_STRING, "Preference"   , MWOPTION.PREFERENCE ) ;
      menuroot.addMenu(             MFT_STRING, "Help"         , MWHELP.ROOT   ) ;
      menuroot.addItem( "Help",     MFT_STRING, "OperateHelp"  , MWHELP.OPHELP ) ;
      menuroot.addItem( "Help",     MFT_STRING, "VersionInfo"  , MWHELP.VERINFO ) ;
      setMenu(menuroot) ;
      
      statusbar = new StatusBar() ;
      statusbar.addStyle(CCS_BOTTOM | SBARS_SIZEGRIP) ;
      addStatusBar(statusbar, "") ;
      
      mainpanel = new MainPanel("MainPanel","MainPanel") ;
      mainpanel.setPosition(0,0,200,200) ;
      mainpanel.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      addObject( mainpanel, null, true, true ) ;

      initSystem() ;
      setWindowTitle() ;
      break ;
      
    case WM_SYSKEYDOWN:
    case WM_KEYDOWN:
    case WM_CHAR:
      sendmsg(mainpanel.oppanel.getHWND(), msg, wp, lp) ;
      break ;

    case WM_INITMENU:
      menuroot.enableItem( "HGcntl",   MWVIEW.HGCP|MWVIEWSUB.SHOW,  hairdb.getHideHGvtx()  ) ;
      menuroot.enableItem( "HGcntl",   MWVIEW.HGCP|MWVIEWSUB.HIDE, !hairdb.getHideHGvtx()  ) ;
      menuroot.enableItem( "HairRoot", MWVIEW.HR|MWVIEWSUB.SHOW,    hairdb.getHideHRvtx()  ) ;
      menuroot.enableItem( "HairRoot", MWVIEW.HR|MWVIEWSUB.HIDE,   !hairdb.getHideHRvtx()  ) ;
      menuroot.enableItem( "Object",   MWVIEW.OBJ|MWVIEWSUB.SHOW,   hairdb.getHideHRface() ) ;
      menuroot.enableItem( "Object",   MWVIEW.OBJ|MWVIEWSUB.HIDE,  !hairdb.getHideHRface() ) ;
      break ;
      
    case WM_CLOSE:
      if( hasChangedDB()==true ){
	if( querySaveDB(hwnd,
			"Do you want to save your change before quitting ?",
			"Not been saved. Are you sure you want to quit ?" )==0 ) break ;
      }
      return( DefWindowProc(hwnd , msg , wp , lp) );
      break ;

    case WM_COMMAND:
      int cmd=LOWORD(wp) ;
      switch(cmd){
      case MWFILE.OPEN :
	if( hasChangedDB()==true ){
	  if( querySaveDB(hwnd,
			  "Do you want to save your change before opening ?",
			  "Not been saved. Are you sure you want to open ?" )==0 ) break ;
	}
	auto   fdlog = new FileDialog() ;
	string fname = fdlog.getLoadFileName(hairfilter,hwnd) ;
	if( fname.length!=0 ){
	  auto hairdb_new = new HairDB(4) ;
	  int rc=hairdb_new.read_hair(fname) ;
	  if( rc==0 ){
	    hairdb=hairdb_new ;
	    reloadDB() ;
	    setWindowTitle() ;
	    opman.execCommand(OPMCMD.REDRAW) ;
	    undobuf.push(new UndoPack(hairdb.cloneHairRootDB(), hairdb.cloneHairGuideDB()),false) ;
	  }else{
	    auto msgbox = new MsgBox(hwnd, hairdb_new.getErrorMsg(), "Error Message", MB_OK ) ;
	    msgbox.create() ;
	  }
	}
	break ;
      
      case MWFILE.SAVE :
	if( hairdb.isObjExists()==false ){
	  auto msgbox = new MsgBox(hwnd, "No need to save.", "Message", MB_OK ) ;
	  msgbox.create() ;
	}else{
	  if( hasChangedDB()==true ){
	    string fname = hairdb.getCurFilename() ;
	    if( fname.length!=0 ){
	      hairdb.write_hair(fname) ;
	      undobuf.setSavePoint() ;
	      setWindowTitle() ;
	    }
	  }
	}
	break ;

      case MWFILE.SAVEAS :
	if( hairdb.isObjExists()==false ){
	  auto msgbox = new MsgBox(hwnd, "No need to save.", "Message", MB_OK ) ;
	  msgbox.create() ;
	}else{
	  auto fdlog = new FileDialog() ;
	  string fname = fdlog.getSaveFileName(hairfilter,hwnd,OFN_OVERWRITEPROMPT,hairdb.getCurFilename()) ;
	  if( fname.length!=0 ){
	    hairdb.write_hair(fname) ;
	    undobuf.setSavePoint() ;
	    setWindowTitle() ;
	  }
	}
	break ;
      
      case MWFILE.IMPORT :
	if( hasChangedDB()==true ){
	  if( querySaveDB(hwnd,
			  "Do you want to save your change before importing ?",
			  "Not been saved. Are you sure you want to import ?" )==0 ) break ;
	}
	auto   fdlog = new FileDialog() ;
	string fname = fdlog.getLoadFileName(objfilter,hwnd) ;
	if( fname.length!=0 ){
	  auto hairdb_new = new HairDB(4) ;
	  int rc=hairdb_new.read_obj(fname) ;
	  if( rc==0 ){
	    hairdb=hairdb_new ;
	    reloadDB() ;
	    setWindowTitle() ;
	    opman.execCommand(OPMCMD.REDRAW) ;
	    undobuf.push(new UndoPack(hairdb.cloneHairRootDB(), hairdb.cloneHairGuideDB()),true) ;
	  }else{
	    auto msgbox = new MsgBox(hwnd, hairdb_new.getErrorMsg(), "Error Message", MB_OK ) ;
	    msgbox.create() ;
	  }
	}
	break ;
      
      case MWFILE.EXPORT :
	RECT wrect ;
        GetWindowRect(getMainWinHWND(),&wrect) ;
	auto exPOVdlog = new ExportPOVDialog("Export POV parameter","editpovraram" ) ;
	exPOVdlog.setDefaultPOVparam(hairdb.getPOVparam()) ;
	exPOVdlog.Create( wrect.left+100,wrect.top+100,hwnd,null ) ;
	if( exPOVdlog.getDlogResult()==1 ){
	  auto fdlog = new FileDialog() ;
	  string fname = fdlog.getSaveFileName(povfilter,hwnd, OFN_OVERWRITEPROMPT) ;
	  if( fname.length!=0 ){
	    hairdb.write_pov(fname,exPOVdlog.getPOVparam()) ;
	  }
	}
	SetFocus(getMainWinHWND()) ;
	break ;
      
      case MWFILE.EXIT :
	sendmsg(WM_CLOSE, 0,0) ;
	break ;

      case MWEDIT.SELALL :
	opman.execCommand(OPMCMD.SELALL) ;
	break ;
      case MWEDIT.UNSELALL :
	opman.execCommand(OPMCMD.UNSELALL) ;
	break ;
	
      case MWVIEW.RESET :
	opman.execCommand(OPMCMD.VIEWRESET) ;
	break ;

      case MWVIEW.HGCP|MWVIEWSUB.SHOW :
      case MWVIEW.HGCP|MWVIEWSUB.HIDE :
	opman.execCommand(OPMCMD.SET_HIDE_HGCP, (cmd & MWVIEWSUB.HIDE)==MWVIEWSUB.HIDE) ;
	break ;

      case MWVIEW.HR|MWVIEWSUB.SHOW :
      case MWVIEW.HR|MWVIEWSUB.HIDE :
	opman.execCommand(OPMCMD.SET_HIDE_HR, (cmd & MWVIEWSUB.HIDE)==MWVIEWSUB.HIDE) ;
	break ;

      case MWVIEW.OBJ|MWVIEWSUB.SHOW :
      case MWVIEW.OBJ|MWVIEWSUB.HIDE :
	opman.execCommand(OPMCMD.SET_HIDE_OBJ, (cmd & MWVIEWSUB.HIDE)==MWVIEWSUB.HIDE) ;
	break ;

      case MWOPTION.PREFERENCE :
	RECT wrect ;
        GetWindowRect(getMainWinHWND(),&wrect) ;
	auto prefdlog = new PrefDialog("Preference Dialog","preferencedialog" ) ;
	prefdlog.setDefaultWorkSpaceEnv(mainpanel.openglframe.getWorkSpaceEnv()) ;
	prefdlog.Create( wrect.left+100,wrect.top+100,hwnd,null ) ;
	if( prefdlog.getDlogResult()==1 ){
	  mainpanel.openglframe.setWorkSpaceEnv(prefdlog.getWorkSpaceEnv()) ;
	  opman.execCommand(OPMCMD.REDRAW) ;
	}
	//SetFocus(getMainWinHWND()) ;
	opman.execCommand(OPMCMD.REDRAW) ; //OpenGLwindow Redraw Bug? workaround
	break ;
	
      case MWHELP.OPHELP :
	string[] helpstr = [ "---------WorkSpace View controls------------",
			     " ",
			     "A+MLd   : Rotate Object                    ",
			     "A+C+MLd : ZoomIn/Out Object                ",
			     "A+S+MLd : Shift Object                     ",
			     " ",         
			     "A=Alt-key                                  ",
			     "C=Control-key                              ",
			     "S=Shift-key                                ",
			     "MLd=Mouse'L'button-drag                    ",
			     " ",
			     "---------Shortcuts-------------------------",
			     " ",
			     " 'a' : select all                          ",
			     " 'g' : generate hairguide                  ",
			     " 'h' : toggle hide object                  ",
			     " 'H' : toggle hide hairguide control point ",
			     " 'r' : reset view                          ",
			     " 'x' : delete surface                      ",
			     " 'y' : generate surface                    ",
			     " SPC : unselect all                        ",
			     " DEL : delete hairguide                    "
			     ] ;
	RECT wrect ;
        GetWindowRect(getMainWinHWND(),&wrect) ;
	auto ophelpdlog = new OpHelpDialog("Operation Help","ophelp" ) ;
	ophelpdlog.setHelpString(helpstr) ;
	ophelpdlog.Create( wrect.left+100,wrect.top+100,300,360,hwnd,null ) ;
	break ;

      case MWHELP.VERINFO :
	string[] verstr = [ "HairMaker for Windows Ver0.10",
			    " ",
			    "programmed by TANE 2004-2013" ] ;
	RECT wrect ;
        GetWindowRect(getMainWinHWND(),&wrect) ;
	auto verinfodlog = new VerInfoDialog("Version Information","verionfodialog" ) ;
	verinfodlog.setVersionString(verstr) ;
	verinfodlog.setHICON(winc.hIcon) ;
	verinfodlog.Create( wrect.left+100,wrect.top+100,hwnd,null ) ;
	break ;

      default :
	break ;
      }
      break;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int querySaveDB(HWND hwnd, string msg1, string msg2){
    auto msgbox = new MsgBox(hwnd, msg1, "", MB_YESNOCANCEL ) ;
    int msgbox_ret = msgbox.create() ;
    if( msgbox_ret==IDYES || msgbox_ret==IDNO ){
      if( msgbox_ret==IDYES ){
	while(1){
	  auto fdlog = new FileDialog() ;
	  string fname = fdlog.getSaveFileName(hairfilter,hwnd,OFN_OVERWRITEPROMPT,hairdb.getCurFilename()) ;
	  if( fname.length!=0 ){
	    hairdb.write_hair(fname) ;
	    setWindowTitle() ;
	  }else{
	    auto remsg = new MsgBox(hwnd, msg2, "", MB_YESNO ) ;
	    int remsg_ret = remsg.create() ;
	    if( remsg_ret==IDYES ) break ;
	  }
	}
      }
      return(1) ;
    }
    return(0) ;
  }    

  int initSystem(){
    if( mainpanel !is null ){
      hairdb = new HairDB(4) ;
      opman  = new OpManager() ;
      undobuf= new UndoBuffer(21) ;
      reloadDB() ;
      return(0) ;
    }
    throw new Exception("Initialize failed.") ;
  }

  int reloadDB(){
    undobuf.clear() ;
    opman.setHairDB(hairdb) ;
    opman.setUndoBuffer(undobuf) ;
    opman.setOpenGLFrame(mainpanel.openglframe) ;
    opman.setOpPanel(mainpanel.oppanel) ;
    opman.setMainWindow(this) ;
    mainpanel.openglframe.setOpManager(opman) ;
    mainpanel.oppanel.setOpManager(opman) ;
    mainpanel.oppanel.setUndoBuffer(undobuf) ;
    mainpanel.oppanel.gidoppanel.setOpManager(opman) ;
    mainpanel.oppanel.gidoppanel.resyncHairDB() ;
    mainpanel.oppanel.syncButtonState() ;
    core.memory.GC.minimize() ;
    return(0) ;
  }

  void setWindowTitle(){
    if( hairdb !is null ){
      string fmt = hasChangedDB()==true ?  "%s %s(*)" : "%s %s" ;
      setWindowText( format(fmt,appname,hairdb.getCurFilename()) ) ;
    }
  }

  bool hasChangedDB(){
    return(undobuf.hasChangedData()) ;
  }
}
