import win32.windows;
import std.stdio;
import std.string;
import std.utf;
import win32frame.writek;
import win32frame.MsgTranslator;

shared static WindowBase[HWND] _WindowList ;
shared static WNDPROC[HWND]    _DefProc    ;
shared static bool ThemeSupported = false ;
shared static bool checkedThemeSupport =false ;
shared static bool disableThemeSupport =false ;

class WindowBase {
  protected WNDCLASSEX  winc;
  private   HWND        hwnd;
  private   string      name ;
  private   HMENU       hmenu ;
  private   bool        finish_create ;
  
  this( string name ){
    this.winc.cbSize      = WNDCLASSEX.sizeof;
    this.winc.hInstance   = GetModuleHandle(null) ;
    this.winc.lpfnWndProc = cast(WNDPROC)(&_WndProc) ;
    this.hwnd             = null ;
    this.name             = name ;
    this.finish_create    = false;
  }

  int setWNDCLASSEX( WNDCLASSEX winc){
    this.winc=winc ;
    return(0) ;
  }

  int registwinc(){
    if( RegisterClassEx(&this.winc)==0 ) return(-1) ;
    return(0) ;
  }
  
  int unregistwinc(){
    if( UnregisterClass(this.winc.lpszClassName,null)==0 ) return(-1) ;
    return(0) ;
  }
  
  int lowcreate(DWORD exstyle, DWORD style,
		int x, int y, int width, int height, HWND phwnd, HMENU hmenu){
    
    if( this.finish_create==false ){
      HWND rethwnd ;
      this.hwnd=null ;
      rethwnd = CreateWindowEx(exstyle,
			       this.winc.lpszClassName,
			       toUTFz!(wchar*)(this.name),
			       style,
			       x,y,width,height,
			       phwnd, hmenu, this.winc.hInstance , cast(LPVOID)(this)
			       );
      if( rethwnd == null ){
	writek("create failed %s %d,%d,%d,%d\n",name, x,y,width,height) ;
	return(-2) ;
      }else{
	this.hwnd=rethwnd ;
      }
      _WindowList[this.hwnd]=cast(shared WindowBase)(this) ;
      _DefProc[this.hwnd]   =cast(shared WNDPROC)(GetWindowLongPtr(this.hwnd , GWL_WNDPROC)) ;
      SetWindowLongPtr( this.hwnd, GWL_WNDPROC, cast(LONG)(cast(size_t*)(&_WndSubProc))) ;
      this.finish_create=true ;
      this.hmenu=hmenu ;
      writek("create hwnd=%08x name=%s\n",cast(int)(this.hwnd),name) ;
      return(0) ;
    }
    return(-1) ;
  }

  HWND getHWND(){
    return( this.hwnd ) ;
  }
  
  HWND getParentHWND(){
    return( GetParent(this.hwnd) ) ;
  }
  
  HINSTANCE getHInstance(){
    return( this.winc.hInstance ) ;
  }

  HMENU getHMENU(){
    return( this.hmenu ) ;
  }

  string getName(){
    return( this.name ) ;
  }

  int setFont(HFONT hfont){
    if( this.hwnd!=null ){
      SendMessage( this.hwnd, WM_SETFONT, cast(WPARAM)(hfont), cast(LPARAM)(TRUE) );
      return(0) ;
    }
    return(-1) ;
  }

  LRESULT sendmsg(UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(this.hwnd,msg,wp,lp)) ;
  }

  LRESULT sendmsg(HWND tgthwnd, UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(tgthwnd,msg,wp,lp)) ;
  }

  LRESULT sendmsgtoparent(UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(this.getParentHWND(),msg,wp,lp)) ;
  }

  BOOL setWindowText(string str){
    int res = cast(int)SendMessage( this.hwnd, WM_SETTEXT, 0, cast(LPARAM)(toUTFz!(wchar*)(str)) ) ;
    if( res<=0 ){
      return(false) ;
    }
    return(true) ;
  }

  wstring getWindowTextUTF16(){
    wchar[] wstr   ; wstr.length = GetWindowTextLength(this.hwnd)+1 ;
    GetWindowText(this.hwnd , wstr.ptr , cast(int)(wstr.length));
    wstr.length=wstr.length-1 ;
    return( std.utf.toUTF16(wstr) ) ;
  }
  
  string getWindowText(){
    wchar[] wstr   ; wstr.length = GetWindowTextLength(this.hwnd)+1 ;
    GetWindowText(this.hwnd , wstr.ptr , cast(int)(wstr.length));
    wstr.length=wstr.length-1 ;
    return( std.utf.toUTF8(wstr.idup) ) ;
  }
  
  LRESULT preproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return( this.proc(hwnd , msg , wp , lp) ) ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return( DefWindowProc(hwnd , msg , wp , lp) ) ;
  }

  LRESULT subproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return(-1) ;
  }

  LRESULT presubproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return(-1) ;
  }

  bool isThemeSupported(){
    return( IsThemeSupported() ) ;
  }

  void disableThemeSupport(){
    DisableThemeSupport() ;
  }

  ~this(){
    writek("----WindowBase called destructor hwnd=%08x name=%s\n",this.hwnd,this.name) ;
    if( this.hwnd!=null ){
      DestroyWindow(this.hwnd);
    }
  }
}

extern (Windows) LRESULT _WndProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  version(ENBMSGTRACE){
    writek("_WndProc : hwnd=%08x, msg=%s\n",cast(int)(hwnd),MsgTranslate(msg)) ;
  }
  if( msg==WM_NCCREATE ){
    synchronized{
      WindowBase wb = cast(WindowBase)(*(cast(size_t*)(lp))) ;
      wb.hwnd=hwnd ;
      _WindowList[hwnd]=cast(shared WindowBase)(wb) ;
    }
  }
  
  if( (hwnd in _WindowList) != null ){
    WindowBase wb=cast(WindowBase)(_WindowList[hwnd]) ;
    LRESULT   res=wb.preproc(hwnd,msg,wp,lp) ;
    if( msg==WM_NCDESTROY ){
      synchronized{
	_WindowList[hwnd]=null ;
	_WindowList.remove(hwnd) ;
	_DefProc[hwnd]=null ;
	_DefProc.remove(hwnd) ;
	wb.hwnd=null ;
      }
    }
    return( res ) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}

extern (Windows) LRESULT _WndSubProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  version(ENBMSGTRACE){
    writek("_WndSubProc : hwnd=%08x, msg=%s\n",cast(int)(hwnd),MsgTranslate(msg)) ;
  }
  if( (hwnd in _DefProc) != null ){
    WindowBase wb=cast(WindowBase)(_WindowList[hwnd]) ;
    LRESULT ret=wb.presubproc(hwnd,msg,wp,lp) ;
    if( ret==0 ) return(0) ;
    WNDPROC wproc = cast(WNDPROC)(_DefProc[hwnd]) ;
    return(CallWindowProc( wproc , hwnd, msg, wp, lp)) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}

bool IsThemeSupported()
{
  if( disableThemeSupport==false ){
    if( checkedThemeSupport == false ){
      HINSTANCE htdll = LoadLibrary("uxtheme.dll") ;
      if( htdll != null ){
	ThemeSupported = true ;
	FreeLibrary(htdll) ;
      }
      checkedThemeSupport=true ;
    }
  }
  return( ThemeSupported ) ;
}

void DisableThemeSupport()
{
  disableThemeSupport=true ;
}
